#!/usr/bin/python3
"""
 @file
 @brief This file verifies all translations are correctly formatted and have the correct # of string replacements
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import argparse
import ast
import shutil
import os
import re
import fnmatch
import subprocess
import sys
from PyQt5.QtCore import QTranslator, QCoreApplication  # type: ignore
from collections import Counter
from typing import Any, Dict, List, Optional, Tuple


# Absolute path of the translations directory
LANG_PATH = os.path.dirname(os.path.abspath(__file__))
REPO_PATH = os.path.dirname(os.path.dirname(LANG_PATH))
DOC_LOCALE_PATH = os.path.join(REPO_PATH, 'doc', 'locale')

# Match '%(name)x' format placeholders
TAG_RE = re.compile(r'%\(([^\)]*)\)(.)')
DOC_SUB_RE = re.compile(r'\|[A-Za-z0-9_]+\|')
PRINTF_TOKEN_RE = re.compile(
    r'%(?:'
    r'%'  # escaped percent literal
    r'|\(([^)]+)\)[#0\-+]*\d*(?:\.\d+)?[diouxXeEfFgGcrsa]'  # named placeholder
    r'|[#0\-+]*\d*(?:\.\d+)?[diouxXeEfFgGcrsa]'  # positional/unnamed placeholder
    r')'
)

# Match numbers in strings (for detecting number-embedded strings like "Line 1")
NUMBER_RE = re.compile(r'\d+')


class BadTranslationsError(Exception):
    pass


class POEntry:
    """Minimal PO entry representation for doc validation."""

    def __init__(self) -> None:
        self.msgid = ""
        self.msgid_plural = ""
        self.msgstrs: Dict[int, str] = {}
        self.references: List[str] = []


def get_doc_locale_from_path(path: str) -> Optional[str]:
    """Extract the doc locale code from a PO file path."""
    parts = os.path.normpath(path).split(os.sep)
    try:
        locale_index = parts.index('locale')
    except ValueError:
        return None
    if locale_index + 1 >= len(parts):
        return None
    return parts[locale_index + 1]


class Color:
    """Color for message output"""
    _red = '\u001b[31m'
    _yellow = '\u001b[33m'
    _green = '\u001b[32m'
    _reset = '\u001b[0m'

    @classmethod
    def red(cls, *args: Optional[Any]) -> str:
        return cls._red + str(*args) + cls._reset

    @classmethod
    def yellow(cls, *args: Optional[Any]) -> str:
        return cls._yellow + str(*args) + cls._reset

    @classmethod
    def green(cls, *args: Optional[Any]) -> str:
        return cls._green + str(*args) + cls._reset


def get_app() -> QCoreApplication:
    """Create the translation app lazily so test discovery doesn't lock in QCoreApplication."""
    app = QCoreApplication.instance()
    if app is not None:
        return app
    return QCoreApplication(sys.argv)


def build_stringlists() -> Dict[str, List]:
    """ Create a dict containing lists of strings, keyed on source filename"""
    all_strings = {}

    for pot_file in [
        'OpenShot.pot',
        'OpenShot_transitions.pot',
        'OpenShot_blender.pot',
        'OpenShot_emojis.pot',
    ]:
        with open(os.path.join(LANG_PATH, 'OpenShot', pot_file)) as f:
            data = f.read()
        all_strings.update({
            pot_file: re.findall('^msgid \"(.*)\"', data, re.MULTILINE)
        })
    return all_strings


def check_trans(strings: List) -> List[Tuple[str, str]]:
    """Check all strings in a list against a given .qm file"""
    app = get_app()
    # Test translation of all strings
    translations = {
        source: app.translate("", source)
        for source in strings
    }
    # Check for replacements with mismatched number of % escapes
    errors = {}
    for s, t in translations.items():
        # Count placeholders in source and translation
        s_count = s.count('%s') + s.count('%d') + s.count('%f')
        t_count = t.count('%s') + t.count('%d') + t.count('%f')

        # If counts match, no error
        if s.count('%s') == t.count('%s') and \
           s.count('%d') == t.count('%d') and \
           s.count('%f') == t.count('%f'):
            continue

        # Special case: source has embedded numbers but no placeholders,
        # and translation has %s placeholders instead.
        # This is valid when translators use a placeholder pattern
        # (e.g., "Line 1" translated as "Linia %s" where %s = "1")
        source_numbers = NUMBER_RE.findall(s)
        if source_numbers and s_count == 0 and t_count == len(source_numbers):
            continue

        errors[s] = t
    # Check for missing/added variable names
    # e.g.: "%(clip_id)s %(value)d" changed to "%(clip)s %(value)d"
    # or mismatched types
    # e.g.: "%(seconds)s" changed to "%(seconds)d"
    named_variables = {
        s: (TAG_RE.findall(s), TAG_RE.findall(t))
        for s, t in translations.items()
        if s.count('%(') > 0
    }
    errors.update({
        s: translations[s]
        for s, (s_vars, t_vars) in named_variables.items()
        # Ensure that t_vars is a strict subset of s_vars
        if not set(t_vars) <= set(s_vars)
    })
    return list(errors.items())


def process_qm(file: str, stringlists: Dict[str, List[str]]) -> int:
    """Scan a translation file against all provided strings"""
    app = get_app()
    # Attempt to load translation file
    basename = os.path.splitext(file)[0]
    translator = QTranslator(app)
    if not translator.load(basename, LANG_PATH):
        print(Color.red('QTranslator failed to load') + f' {file}')
        return 1

    app.installTranslator(translator)

    # Build a dict mapping source POTfiles to lists of error pairs
    error_sets = {
        sourcefile: check_trans(strings)
        for sourcefile, strings in stringlists.items()
    }

    app.removeTranslator(translator)

    # Display any errors found, grouped by source POT file
    error_count = sum([len(v) for v in error_sets.values()])
    invalid_msg = "Invalid"
    if error_count:
        print(f'{file}: ' + Color.red(f'{error_count} total errors'))
    for pot, errset in error_sets.items():
        if not errset:
            continue
        width = len(pot)
        for source, trans in errset:
            print(Color.yellow(f'{pot}:') + f' {source}')
            print(Color.red(f'{invalid_msg:>{width}}:') + f' {trans}\n')
    return error_count


def scan_all_qm(filenames: List[str] = None) -> None:
    all_strings = build_stringlists()
    if not filenames:
        filenames = fnmatch.filter(os.listdir(LANG_PATH), 'OpenShot*.qm')
    # Loop through language files and count errors
    total_errors = sum([
        process_qm(filename, all_strings)
        for filename in filenames
    ])

    string_count = sum([len(s) for s in all_strings.values()])
    lang_count = len(filenames)

    print(f"Tested {Color.yellow(string_count)} strings on "
          + f"{Color.yellow(lang_count)} translation files.")
    if total_errors > 0:
        raise BadTranslationsError(f"Found {total_errors} translation errors")


def po_unquote(text: str) -> str:
    """Decode a quoted PO string token."""
    return ast.literal_eval(text)


def extract_percent_tokens(text: str) -> Tuple[Counter, Counter]:
    """Extract printf-style percent tokens."""
    tokens = Counter()
    named_tokens = Counter()

    for match in PRINTF_TOKEN_RE.finditer(text):
        token = match.group(0)
        tokens[token] += 1
        named_match = TAG_RE.match(token)
        if named_match:
            named_tokens[named_match.groups()] += 1

    return tokens, named_tokens


def extract_doc_substitutions(text: str) -> Counter:
    """Extract Sphinx substitution tokens like |icon_name|."""
    return Counter(DOC_SUB_RE.findall(text))


def flush_po_entry(entries: List[POEntry], entry: POEntry) -> POEntry:
    """Append the current PO entry if it contains any data, then start a new one."""
    if entry.msgid or entry.msgstrs or entry.msgid_plural or entry.references:
        entries.append(entry)
    return POEntry()


def parse_po_file(path: str) -> List[POEntry]:
    """Parse enough of a PO file to inspect msgid/msgstr tokens."""
    entries: List[POEntry] = []
    entry = POEntry()
    active_field: Optional[Tuple[str, Optional[int]]] = None

    with open(path, encoding='utf-8') as handle:
        for raw_line in handle:
            line = raw_line.rstrip('\n')

            if not line.strip():
                entry = flush_po_entry(entries, entry)
                active_field = None
                continue

            if line.startswith('#~'):
                continue

            if line.startswith('#:'):
                entry.references.extend(line[2:].strip().split())
                continue

            if line.startswith('#'):
                continue

            if line.startswith('msgid_plural '):
                entry.msgid_plural = po_unquote(line[len('msgid_plural '):])
                active_field = ('msgid_plural', None)
                continue

            if line.startswith('msgid '):
                if entry.msgid or entry.msgstrs or entry.msgid_plural:
                    entry = flush_po_entry(entries, entry)
                entry.msgid = po_unquote(line[len('msgid '):])
                active_field = ('msgid', None)
                continue

            if line.startswith('msgstr['):
                index_text, value = line.split(']', 1)
                index = int(index_text[len('msgstr['):])
                entry.msgstrs[index] = po_unquote(value.strip()[1:])
                active_field = ('msgstr', index)
                continue

            if line.startswith('msgstr '):
                entry.msgstrs[0] = po_unquote(line[len('msgstr '):])
                active_field = ('msgstr', 0)
                continue

            if line.startswith('msgctxt '):
                active_field = ('msgctxt', None)
                continue

            if line.startswith('"'):
                value = po_unquote(line)
                if active_field == ('msgid', None):
                    entry.msgid += value
                elif active_field == ('msgid_plural', None):
                    entry.msgid_plural += value
                elif active_field and active_field[0] == 'msgstr' and active_field[1] is not None:
                    entry.msgstrs[active_field[1]] = entry.msgstrs.get(active_field[1], '') + value
                continue

    flush_po_entry(entries, entry)
    return entries


def validate_doc_entry(path: str, entry: POEntry) -> List[str]:
    """Check doc translation entries for placeholder and substitution preservation."""
    if entry.msgid == "":
        header = entry.msgstrs.get(0, '')
        header_fields: Dict[str, str] = {}
        for line in header.splitlines():
            if ':' not in line:
                continue
            key, value = line.split(':', 1)
            header_fields[key.strip()] = value.strip()

        expected_language = get_doc_locale_from_path(path)
        actual_language = header_fields.get('Language', '')
        if not expected_language:
            return []

        errors = []
        if not actual_language:
            errors.append(f"{path}: PO header is missing a Language value")
        elif actual_language != expected_language:
            errors.append(
                f"{path}: PO header Language '{actual_language}' does not match locale '{expected_language}'"
            )
        return errors

    source_strings = [entry.msgid]
    if entry.msgid_plural:
        source_strings.append(entry.msgid_plural)

    expected_named = Counter()
    expected_tokens = Counter()
    expected_subs = Counter()
    for source in source_strings:
        source_tokens, source_named = extract_percent_tokens(source)
        expected_named |= source_named
        expected_tokens |= source_tokens
        expected_subs |= extract_doc_substitutions(source)

    errors = []
    for plural_index, translation in sorted(entry.msgstrs.items()):
        if not translation.strip():
            continue

        actual_tokens, actual_named = extract_percent_tokens(translation)
        actual_subs = extract_doc_substitutions(translation)

        label = path
        if entry.references:
            label += f" ({', '.join(entry.references[:2])})"
        if len(entry.msgstrs) > 1:
            label += f" [msgstr[{plural_index}]]"

        if actual_named != expected_named:
            errors.append(
                f"{label}: named placeholders differ\n"
                f"  msgid: {entry.msgid}\n"
                f"  msgstr: {translation}"
            )

        if actual_tokens != expected_tokens:
            errors.append(
                f"{label}: printf tokens differ\n"
                f"  msgid: {entry.msgid}\n"
                f"  msgstr: {translation}"
            )

        if actual_subs != expected_subs:
            errors.append(
                f"{label}: Sphinx substitution tokens differ\n"
                f"  msgid: {entry.msgid}\n"
                f"  msgstr: {translation}"
            )

    return errors


def process_doc_po(path: str) -> int:
    """Run syntax and token-preservation checks on a doc PO file."""
    errors: List[str] = []
    msgfmt_path = shutil.which('msgfmt')
    trusted_path = os.path.abspath(path)
    trusted_roots = (os.path.abspath(LANG_PATH), os.path.abspath(DOC_LOCALE_PATH))

    if not msgfmt_path:
        errors.append(f"{path}: msgfmt executable not found")
        return len(errors)

    if not any(
        os.path.commonpath([trusted_path, root]) == root
        for root in trusted_roots
    ):
        errors.append(f"{path}: path is outside trusted locale roots")
        return len(errors)

    msgfmt_args = [
        msgfmt_path,
        '--check',
        '--check-format',
        '--output-file=/dev/null',
        trusted_path,
    ]
    msgfmt = subprocess.run(
        msgfmt_args,
        capture_output=True,
        text=True,
        check=False,
    )
    if msgfmt.returncode != 0:
        stderr = msgfmt.stderr.strip() or msgfmt.stdout.strip() or 'msgfmt validation failed'
        errors.append(f"{path}: {stderr}")

    try:
        entries = parse_po_file(path)
    except Exception as ex:
        errors.append(f"{path}: PO parse failed: {ex}")
        entries = []

    for entry in entries:
        errors.extend(validate_doc_entry(path, entry))

    if errors:
        print(Color.red(f'{path}: {len(errors)} total errors'))
        for error in errors:
            print(error + '\n')
    return len(errors)


def scan_doc_po_files() -> None:
    """Validate all documentation PO files."""
    po_files = []
    for root, _, files in os.walk(DOC_LOCALE_PATH):
        for filename in files:
            if filename.endswith('.po'):
                po_files.append(os.path.join(root, filename))
    po_files.sort()

    total_errors = sum(process_doc_po(path) for path in po_files)
    print(f"Tested {Color.yellow(len(po_files))} documentation PO files.")
    if total_errors > 0:
        raise BadTranslationsError(f"Found {total_errors} documentation translation errors")


def parse_args(argv: List[str]) -> argparse.Namespace:
    """Parse CLI arguments without breaking the existing filename-only behavior."""
    parser = argparse.ArgumentParser(description='Validate OpenShot translation files.')
    parser.add_argument('filenames', nargs='*', help='Optional .qm files to validate')
    parser.add_argument('--docs', action='store_true', help='Validate doc gettext PO files')
    parser.add_argument('--all', action='store_true', help='Validate both app and doc translations')
    return parser.parse_args(argv)


# Autorun if used as script
if __name__ == '__main__':
    try:
        args = parse_args(sys.argv[1:])
        if args.all:
            scan_all_qm(args.filenames)
            scan_doc_po_files()
        elif args.docs:
            scan_doc_po_files()
        else:
            scan_all_qm(args.filenames)
    except BadTranslationsError as ex:
        print(Color.red(str(ex) + "! See above."))
        exit(1)
    else:
        print(Color.green("No errors found."))
        exit(0)



"""
 @file
 @brief Effect badge and drop handling helpers.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from PyQt5.QtCore import QPointF, QRectF, Qt
from classes.query import Clip


class EffectInteractionMixin:
    def _apply_effect_drop(self, effect_names, pos_seconds, track_num, drop_pos=None):
        if not effect_names:
            return
        timeline = getattr(self.win, "timeline", None)
        if not timeline:
            return
        pos_seconds = max(0.0, float(pos_seconds))

        if isinstance(drop_pos, QPointF):
            self.geometry.ensure()
            for rect, clip, _selected in self.geometry.iter_clips(reverse=True):
                if not rect.contains(drop_pos):
                    continue
                data = clip.data if isinstance(clip.data, dict) else {}
                clip_track = data.get("layer")
                try:
                    clip_track = int(clip_track)
                except (TypeError, ValueError):
                    continue
                if self._is_track_locked(clip_track):
                    return
                clip_position = float(data.get("position", 0.0) or 0.0)
                clip_start = float(data.get("start", 0.0) or 0.0)
                clip_end = float(data.get("end", clip_start) or clip_start)
                duration = clip_end - clip_start
                if duration <= 0.0:
                    continue
                clip_finish = clip_position + duration
                target_seconds = min(max(pos_seconds, clip_position), clip_finish)
                timeline.addEffect(effect_names, QPointF(target_seconds, clip_track))
                return

        try:
            track_num = int(track_num)
        except (TypeError, ValueError):
            return
        if self._is_track_locked(track_num):
            return
        candidates = Clip.filter(layer=track_num)
        for clip in candidates:
            data = clip.data if isinstance(clip.data, dict) else {}
            clip_position = float(data.get("position", 0.0) or 0.0)
            clip_start = float(data.get("start", 0.0) or 0.0)
            clip_end = float(data.get("end", clip_start) or clip_start)
            duration = clip_end - clip_start
            if duration <= 0.0:
                continue
            clip_finish = clip_position + duration
            if pos_seconds == 0.0 or clip_position <= pos_seconds <= clip_finish:
                timeline.addEffect(effect_names, QPointF(pos_seconds, track_num))
                break

    def _effect_icon_at(self, pos):
        for entry in reversed(self._effect_icon_rects):
            rect = entry.get("rect")
            if isinstance(rect, QRectF) and rect.contains(pos):
                return entry
        return None

    def _trigger_effect_context_menu(self, icon_entry, modifiers=None):
        """Handle context menu interaction on an effect badge."""
        if not isinstance(icon_entry, dict):
            return False
        effect = icon_entry.get("effect")
        effect_id = icon_entry.get("effect_id")
        if effect_id is None and isinstance(effect, dict):
            effect_id = effect.get("id")
        if effect_id is None:
            return False
        effect_id_str = str(effect_id)
        ctrl = False
        if modifiers is None and self._last_event and hasattr(self._last_event, "modifiers"):
            modifiers = self._last_event.modifiers()
        if modifiers is not None:
            ctrl = bool(modifiers & Qt.ControlModifier)
        self._select_timeline_item(effect_id_str, "effect", not ctrl)
        timeline = getattr(self.win, "timeline", None)
        if timeline:
            timeline.ShowEffectMenu(effect_id_str)
        return True

    def _selected_effect_ids(self):
        selected = getattr(self.win, "selected_effects", [])
        return {str(eff) for eff in selected if eff is not None}

    def _handle_effect_icon_click(self, entry):
        if not isinstance(entry, dict):
            return
        effect = entry.get("effect")
        if not isinstance(effect, dict):
            return
        effect_id = entry.get("effect_id")
        if effect_id is None:
            effect_id = effect.get("id")
        if effect_id is None:
            return
        effect_id_str = str(effect_id)
        modifiers = Qt.NoModifier
        if self._last_event and hasattr(self._last_event, "modifiers"):
            modifiers = self._last_event.modifiers()
        ctrl = bool(modifiers & Qt.ControlModifier)
        self._select_timeline_item(effect_id_str, "effect", not ctrl)



"""
 @file
 @brief This file is used to find missing files when opening or importing a project
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import html
import os
from classes import info
from classes.app import get_app
from classes.path_utils import native_display_path, wrapped_path_html
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QFileDialog, QDialog, QVBoxLayout, QHBoxLayout,
    QPushButton, QLabel
)

# Keep track of all previously checked paths, and keep checking them
known_paths = [info.HOME_PATH]


def _display_directory_html(file_path):
    directory_path = os.path.dirname(file_path)
    if not directory_path:
        directory_path = "."
    return wrapped_path_html(directory_path)


def _deepest_existing_parent(path_value):
    """Return the deepest existing directory in a missing file path."""
    candidate = os.path.abspath(path_value or "")
    if not candidate:
        return ""

    if os.path.isdir(candidate):
        return candidate

    candidate = os.path.dirname(candidate)
    while candidate and not os.path.exists(candidate):
        parent = os.path.dirname(candidate)
        if parent == candidate:
            break
        candidate = parent

    if candidate and os.path.isdir(candidate):
        return candidate
    return ""


def _show_missing_file_dialog(file_name, file_path, starting_dir):
    _ = get_app()._tr
    browse_dir = starting_dir if os.path.isdir(starting_dir) else _deepest_existing_parent(starting_dir)
    if not browse_dir:
        browse_dir = info.HOME_PATH
    dialog = QDialog(None)
    dialog.setWindowTitle(_("Locate Missing Files"))
    dialog.setModal(True)
    dialog.setMinimumSize(420, 130)

    layout = QVBoxLayout(dialog)
    layout.setContentsMargins(12, 10, 12, 10)
    layout.setSpacing(0)

    intro_label = QLabel(_("This file is missing. You can locate it or remove it from the project."))
    intro_label.setObjectName("lblMissingFileHint")
    intro_label.setWordWrap(True)
    layout.addWidget(intro_label)
    layout.addSpacing(12)

    name_label = QLabel(f"<b>{html.escape(file_name)}</b>")
    name_label.setTextFormat(Qt.RichText)
    name_label.setTextInteractionFlags(Qt.TextSelectableByMouse)
    name_label.setContentsMargins(8, 0, 0, 0)
    layout.addWidget(name_label)

    dir_label = QLabel(_display_directory_html(file_path))
    dir_label.setObjectName("lblMissingFilePath")
    dir_label.setWordWrap(True)
    dir_label.setTextFormat(Qt.RichText)
    dir_label.setTextInteractionFlags(Qt.TextSelectableByMouse)
    dir_label.setContentsMargins(8, 0, 0, 0)
    dir_label.setToolTip(native_display_path(os.path.dirname(file_path) or "."))
    layout.addWidget(dir_label)
    layout.addStretch(1)

    button_row = QHBoxLayout()
    button_row.setContentsMargins(0, 0, 0, 0)
    button_row.setSpacing(6)
    skip_all_button = QPushButton(_("Remove All Missing"))
    skip_file_button = QPushButton(_("Remove"))
    browse_button = QPushButton(_("Browse..."))
    browse_button.setObjectName("acceptButton")
    button_row.addWidget(skip_all_button)
    button_row.addStretch(1)
    button_row.addWidget(skip_file_button)
    button_row.addWidget(browse_button)
    layout.addLayout(button_row)

    result = {"action": "file", "path": file_path}

    def pick_browse():
        selected_dir = QFileDialog.getExistingDirectory(
            None,
            _("Locate folder containing: %s") % file_name,
            browse_dir,
        )
        if selected_dir:
            result["action"] = "locate"
            result["path"] = os.path.join(selected_dir, file_name)
            dialog.accept()

    def pick_skip_all():
        result["action"] = "all"
        dialog.reject()

    browse_button.clicked.connect(pick_browse)
    skip_file_button.clicked.connect(dialog.reject)
    skip_all_button.clicked.connect(pick_skip_all)
    browse_button.setDefault(True)
    browse_button.setAutoDefault(True)
    browse_button.setFocus()

    dialog.exec_()
    return result["action"], result["path"]


def find_missing_file(file_path, prompt_state=None):
    """Find a missing file name or file path, and return valid path."""
    _ = get_app()._tr
    modified = False
    skipped = False
    if prompt_state is not None:
        prompt_state["last_skip"] = None

    # If user cancelled prompts, skip searching
    if prompt_state and prompt_state.get("cancelled"):
        if prompt_state is not None:
            prompt_state["last_skip"] = "all"
        return ("", modified, True)

    # Bail if path is already valid
    if os.path.exists(file_path):
        return (file_path, modified, skipped)

    # Original filename
    file_name = os.path.split(file_path)[-1]

    # Loop through all known paths, and check for this file
    for known_path in known_paths:
        possible_path = os.path.join(known_path, file_name)
        if os.path.exists(possible_path):
            modified = True
            return (possible_path, modified, skipped)

    # Check if path exists
    while not os.path.exists(file_path):
        recommended_path = _deepest_existing_parent(file_path)
        if not recommended_path:
            recommended_path = get_app().project.current_filepath or ""
            if not recommended_path:
                recommended_path = info.HOME_PATH
            else:
                recommended_path = os.path.dirname(recommended_path)
        action, selected_path = _show_missing_file_dialog(file_name, file_path, recommended_path)
        modified = True

        if action == "all":
            # User skipped all missing file prompts
            skipped = True
            if prompt_state is not None:
                prompt_state["cancelled"] = True
                prompt_state["last_skip"] = "all"
            return ("", modified, skipped)

        if action == "file":
            # User skipped this missing file only
            skipped = True
            if prompt_state is not None:
                prompt_state["last_skip"] = "file"
            return ("", modified, skipped)

        file_path = selected_path
        folder_to_check = os.path.dirname(file_path)
        if folder_to_check and folder_to_check not in known_paths:
            known_paths.append(folder_to_check)

    # Return found file_path
    return (file_path, modified, skipped)



"""
 @file
 @brief This file contains the Generate media dialog.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import json
import functools

from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QIcon, QPixmap, QColor
from PyQt5.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout, QFormLayout, QLabel, QLineEdit,
    QComboBox, QTextEdit, QTabWidget, QWidget, QPushButton, QMessageBox,
    QDoubleSpinBox, QSpinBox
)

from classes import info
from classes.logger import log
from classes.thumbnail import GetThumbPath
from classes.query import File
from windows.region import SelectRegion
from windows.color_picker import ColorPicker


class GenerateMediaDialog(QDialog):
    """Minimal generate dialog with a simple default-first layout."""

    PREVIEW_WIDTH = 180
    PREVIEW_HEIGHT = 128

    def __init__(
        self,
        source_file=None,
        templates=None,
        preselected_template_id=None,
        dialog_title=None,
        parent=None,
    ):
        super().__init__(parent)
        self.source_file = source_file
        self.templates = templates or []
        self.template_map = {
            str(t.get("id", "")).strip(): (t.get("template") or {})
            for t in self.templates
        }
        self.preselected_template_id = str(preselected_template_id or "").strip()
        self._coordinates_positive_text = ""
        self._coordinates_negative_text = ""
        self._rectangles_positive_text = ""
        self._rectangles_negative_text = ""
        self._auto_mode = False
        self._tracking_selection_payload = {}
        self.setObjectName("generateDialog")
        self.setWindowTitle(str(dialog_title or "AI Tools"))
        self.setMinimumWidth(620)
        self.setMinimumHeight(460)

        root = QVBoxLayout(self)
        root.setContentsMargins(14, 14, 14, 14)
        root.setSpacing(10)

        root.addLayout(self._build_top_block())

        self.tabs = QTabWidget(self)
        self.tabs.setObjectName("generateTabs")
        self.page_prompt = self._build_prompt_tab()
        self.page_reference = self._build_reference_tab()
        self.page_points = self._build_points_tab()
        self.page_highlight = self._build_highlight_tab()
        self.prompt_tab_index = self.tabs.addTab(self.page_prompt, "Prompt")
        self.reference_tab_index = self.tabs.addTab(self.page_reference, "Reference")
        self.points_tab_index = self.tabs.addTab(self.page_points, "Tracking")
        self.highlight_tab_index = self.tabs.addTab(self.page_highlight, "Highlight")
        root.addWidget(self.tabs, 1)

        button_row = QHBoxLayout()
        button_row.addStretch(1)
        self.cancel_button = QPushButton("Cancel")
        self.generate_button = QPushButton("Generate")
        self.generate_button.setIcon(QIcon(":/icons/Humanity/actions/16/star.svg"))
        self.cancel_button.clicked.connect(self.reject)
        self.generate_button.clicked.connect(self._on_generate_clicked)
        button_row.addWidget(self.cancel_button)
        button_row.addWidget(self.generate_button)
        root.addLayout(button_row)
        self._initialize_dialog_state()

    def _current_coordinates_text(self):
        coordinates_positive = str(self._coordinates_positive_text or "").strip()
        coordinates_negative = str(self._coordinates_negative_text or "").strip()
        rects_positive = str(self._rectangles_positive_text or "").strip()
        rects_negative = str(self._rectangles_negative_text or "").strip()
        auto_mode = bool(self._auto_mode)
        tracking_payload = dict(self._tracking_selection_payload or {})
        if not coordinates_positive and hasattr(self, "points_preview"):
            preview_text = self.points_preview.toPlainText().strip()
            if preview_text.startswith("{"):
                try:
                    payload = json.loads(preview_text.replace("'", "\""))
                    coordinates_positive = str(payload.get("positive", "")).strip() or coordinates_positive
                    coordinates_negative = str(payload.get("negative", "")).strip() or coordinates_negative
                    rects_positive = str(payload.get("positive_rects", "")).strip() or rects_positive
                    rects_negative = str(payload.get("negative_rects", "")).strip() or rects_negative
                    auto_mode = bool(payload.get("auto_mode", auto_mode))
                    if isinstance(payload.get("tracking_selection"), dict):
                        tracking_payload = payload.get("tracking_selection")
                except Exception:
                    pass
        prompt_text = self.prompt_edit.toPlainText().strip()
        return coordinates_positive, coordinates_negative, rects_positive, rects_negative, auto_mode, tracking_payload, prompt_text

    def get_payload(self):
        coordinates_positive, coordinates_negative, rects_positive, rects_negative, auto_mode, tracking_payload, prompt_text = self._current_coordinates_text()
        highlight_color = self.highlight_color.name(QColor.HexArgb) if hasattr(self, "highlight_color") else ""
        border_color = self.border_color.name(QColor.HexArgb) if hasattr(self, "border_color") else ""
        border_width = int(self.border_width_spin.value()) if hasattr(self, "border_width_spin") else 0
        highlight_opacity = float(self.highlight_opacity_spin.value()) if hasattr(self, "highlight_opacity_spin") else 0.0
        mask_brightness = float(self.mask_brightness_spin.value()) if hasattr(self, "mask_brightness_spin") else 1.0
        background_brightness = float(self.background_brightness_spin.value()) if hasattr(self, "background_brightness_spin") else 1.0
        return {
            "name": self.name_edit.text().strip(),
            "template_id": self.template_combo.currentData() or self.template_combo.currentText(),
            "prompt": prompt_text,
            "reference_image_file_id": self.reference_image_combo.currentData() if hasattr(self, "reference_image_combo") else "",
            "coordinates_positive": coordinates_positive,
            "coordinates_negative": coordinates_negative,
            "rectangles_positive": rects_positive,
            "rectangles_negative": rects_negative,
            "auto_mode": bool(auto_mode),
            "tracking_selection": tracking_payload,
            "highlight_color": highlight_color,
            "highlight_opacity": highlight_opacity,
            "border_color": border_color,
            "border_width": border_width,
            "mask_brightness": mask_brightness,
            "background_brightness": background_brightness,
        }

    def _build_top_block(self):
        block = QHBoxLayout()
        block.setSpacing(12)

        if self.source_file:
            self.thumbnail_label = QLabel()
            self.thumbnail_label.setFixedSize(self.PREVIEW_WIDTH, self.PREVIEW_HEIGHT)
            self.thumbnail_label.setAlignment(Qt.AlignCenter)
            self.thumbnail_label.setStyleSheet("border: 1px solid palette(mid);")
            self._load_thumbnail()
            block.addWidget(self.thumbnail_label, 0)

        setup_form = QFormLayout()
        setup_form.setContentsMargins(0, 0, 0, 0)
        setup_form.setVerticalSpacing(8)

        default_name = "generation"
        if self.source_file:
            path = self.source_file.data.get("path", "")
            if path:
                default_name = "{}_gen".format(os.path.splitext(os.path.basename(path))[0])

        self.name_edit = QLineEdit()
        self.name_edit.setPlaceholderText("Output file name")
        self.name_edit.setText(default_name)
        setup_form.addRow("Name", self.name_edit)

        self.template_combo = QComboBox()
        if self.templates:
            for template in self.templates:
                self.template_combo.addItem(template.get("name", ""), template.get("id", ""))
        else:
            self.template_combo.addItem("Basic Text to Image", "txt2img-basic")
        if self.preselected_template_id:
            index = self.template_combo.findData(self.preselected_template_id)
            if index >= 0:
                self.template_combo.setCurrentIndex(index)
        self.template_combo.currentIndexChanged.connect(self._on_template_changed)
        setup_form.addRow("Template", self.template_combo)

        right_container = QWidget(self)
        right_container.setLayout(setup_form)
        block.addWidget(right_container, 1)
        return block

    def _build_reference_tab(self):
        tab = QWidget(self)
        tab.setObjectName("pageReference")
        layout = QVBoxLayout(tab)
        layout.setContentsMargins(8, 8, 8, 8)
        layout.setSpacing(10)

        self.reference_image_combo = QComboBox()
        self.reference_image_combo.setSizeAdjustPolicy(QComboBox.AdjustToMinimumContentsLengthWithIcon)
        self.reference_image_combo.setIconSize(QSize(96, 96))
        self.reference_image_combo.addItem("Choose reference image...", "")
        self._populate_reference_image_combo()
        layout.addWidget(self.reference_image_combo)
        layout.addStretch(1)
        return tab

    def _build_prompt_tab(self):
        tab = QWidget(self)
        tab.setObjectName("pagePrompt")
        layout = QVBoxLayout(tab)
        layout.setContentsMargins(8, 8, 8, 8)
        self.prompt_edit = QTextEdit()
        self.prompt_edit.setPlaceholderText("Prompt (optional)")
        self.prompt_edit.setMinimumHeight(140)
        layout.addWidget(self.prompt_edit)
        return tab

    def _build_points_tab(self):
        tab = QWidget(self)
        tab.setObjectName("pagePoints")
        layout = QVBoxLayout(tab)
        layout.setContentsMargins(8, 8, 8, 8)
        controls = QHBoxLayout()
        self.pick_points_button = QPushButton("Select objects for tracking")
        self.clear_points_button = QPushButton("Clear")
        self.pick_points_button.clicked.connect(self._choose_tracking_clicked)
        self.clear_points_button.clicked.connect(self._clear_points_clicked)
        controls.addWidget(self.pick_points_button)
        controls.addWidget(self.clear_points_button)
        controls.addStretch(1)
        layout.addLayout(controls)

        self.points_preview = QTextEdit()
        self.points_preview.setReadOnly(True)
        self.points_preview.setMinimumHeight(90)
        layout.addWidget(self.points_preview)
        layout.addStretch(1)
        return tab

    def _build_highlight_tab(self):
        tab = QWidget(self)
        tab.setObjectName("pageHighlight")
        layout = QFormLayout(tab)
        layout.setContentsMargins(8, 8, 8, 8)
        layout.setVerticalSpacing(8)
        self.highlight_color = QColor("#2EA6FF")
        self.highlight_color.setAlphaF(0.70)
        self.border_color = QColor("#FFFFFF")
        self.border_color.setAlphaF(1.0)
        self.highlight_color_button = QPushButton("Choose Color")
        self.highlight_color_button.clicked.connect(self._pick_highlight_color)
        self.border_color_button = QPushButton("Choose Color")
        self.border_color_button.clicked.connect(self._pick_border_color)
        self.highlight_opacity_spin = QDoubleSpinBox()
        self.highlight_opacity_spin.setRange(0.0, 1.0)
        self.highlight_opacity_spin.setSingleStep(0.05)
        self.highlight_opacity_spin.setValue(0.28)
        self.border_width_spin = QSpinBox()
        self.border_width_spin.setRange(0, 64)
        self.border_width_spin.setValue(2)
        self.mask_brightness_spin = QDoubleSpinBox()
        self.mask_brightness_spin.setRange(0.0, 3.0)
        self.mask_brightness_spin.setSingleStep(0.05)
        self.mask_brightness_spin.setValue(1.15)
        self.background_brightness_spin = QDoubleSpinBox()
        self.background_brightness_spin.setRange(0.0, 3.0)
        self.background_brightness_spin.setSingleStep(0.05)
        self.background_brightness_spin.setValue(0.75)
        layout.addRow("Background Color", self.highlight_color_button)
        layout.addRow("Background Opacity", self.highlight_opacity_spin)
        layout.addRow("Border Color", self.border_color_button)
        layout.addRow("Border Width", self.border_width_spin)
        layout.addRow("Mask Brightness", self.mask_brightness_spin)
        layout.addRow("Background Brightness", self.background_brightness_spin)
        self._update_highlight_color_button()
        self._update_border_color_button()
        return tab

    @staticmethod
    def _best_contrast(bg):
        colrgb = bg.getRgbF()
        lum = (0.299 * colrgb[0] + 0.587 * colrgb[1] + 0.114 * colrgb[2])
        return QColor(Qt.white) if lum < 0.5 else QColor(Qt.black)

    def _color_callback(self, setter_fn, refresh_fn, color):
        if not color or not color.isValid():
            return
        setter_fn(color)
        refresh_fn()

    def _pick_highlight_color(self):
        callback = functools.partial(
            self._color_callback,
            self._set_highlight_color,
            self._update_highlight_color_button,
        )
        ColorPicker(
            self.highlight_color,
            parent=self,
            title="Select a Color",
            callback=callback,
        )

    def _pick_border_color(self):
        callback = functools.partial(
            self._color_callback,
            self._set_border_color,
            self._update_border_color_button,
        )
        ColorPicker(
            self.border_color,
            parent=self,
            title="Select a Color",
            callback=callback,
        )

    def _set_highlight_color(self, color):
        self.highlight_color = QColor(color)

    def _set_border_color(self, color):
        self.border_color = QColor(color)

    def _update_highlight_color_button(self):
        fg = self._best_contrast(self.highlight_color)
        self.highlight_color_button.setStyleSheet(
            "QPushButton{background-color:%s;color:%s;}" %
            (self.highlight_color.name(QColor.HexArgb), fg.name())
        )

    def _update_border_color_button(self):
        fg = self._best_contrast(self.border_color)
        self.border_color_button.setStyleSheet(
            "QPushButton{background-color:%s;color:%s;}" %
            (self.border_color.name(QColor.HexArgb), fg.name())
        )

    def _load_thumbnail(self):
        path = ""
        media_type = self.source_file.data.get("media_type")
        if media_type in ["video", "image"]:
            path = GetThumbPath(self.source_file.id, 1)
        elif media_type == "audio":
            path = os.path.join(info.PATH, "images", "AudioThumbnail.svg")

        pix = QPixmap(path) if path else QPixmap()
        if not pix.isNull():
            pix = pix.scaled(
                self.PREVIEW_WIDTH - 2,
                self.PREVIEW_HEIGHT - 2,
                Qt.KeepAspectRatio,
                Qt.SmoothTransformation,
            )
            self.thumbnail_label.setPixmap(pix)
        else:
            self.thumbnail_label.setText("No Preview")

    def _on_generate_clicked(self):
        if not self.name_edit.text().strip():
            self.name_edit.setFocus(Qt.TabFocusReason)
            return
        if self._needs_reference_image() and not str(self.reference_image_combo.currentData() or "").strip():
            QMessageBox.warning(
                self,
                "Missing Reference Image",
                "Choose a reference image from Project Files.",
            )
            self.reference_image_combo.setFocus(Qt.TabFocusReason)
            return
        if self._is_track_object_template():
            coordinates_positive, _coordinates_negative, rects_positive, _rects_negative, auto_mode, _tracking_payload, prompt_text = self._current_coordinates_text()
            if (not auto_mode) and (not coordinates_positive) and (not rects_positive) and (not str(prompt_text or "").strip()):
                QMessageBox.warning(
                    self,
                    "Missing Selection",
                    "No SAM2 seed was provided. Add tracking points/rectangles or enter a prompt.",
                )
                self.tabs.setCurrentWidget(self.page_points)
                return
        self.accept()

    def _is_track_object_template(self):
        template_id = str(self.template_combo.currentData() or "").strip().lower()
        return template_id in (
            "video-blur-anything-sam2",
            "video-mask-anything-sam2",
            "video-highlight-anything-sam2",
            "image-blur-anything-sam2",
            "image-mask-anything-sam2",
            "image-highlight-anything-sam2",
        )

    def _is_highlight_template(self):
        template_id = str(self.template_combo.currentData() or "").strip().lower()
        return template_id in ("video-highlight-anything-sam2", "image-highlight-anything-sam2")

    def _current_template(self):
        template_id = str(self.template_combo.currentData() or "").strip()
        return self.template_map.get(template_id, {})

    def _needs_reference_image(self):
        return bool(self._current_template().get("needs_reference_image", False))

    def _on_template_changed(self, index):
        _ = index
        is_track_template = self._is_track_object_template()
        is_highlight_template = self._is_highlight_template()
        needs_inputs = self._needs_reference_image()
        self.reference_image_combo.setVisible(needs_inputs)
        self._set_tab_visible(self.reference_tab_index, needs_inputs)
        self._set_tab_visible(self.prompt_tab_index, is_track_template)
        self._set_tab_visible(self.points_tab_index, is_track_template)
        self._set_tab_visible(self.highlight_tab_index, is_track_template and is_highlight_template)
        self.pick_points_button.setEnabled(bool(self.source_file) and is_track_template)
        self.clear_points_button.setEnabled(is_track_template)
        if is_track_template:
            self.pick_points_button.setText("Select objects for tracking")
            self.tabs.setCurrentWidget(self.page_points)
        else:
            self._set_tab_visible(self.prompt_tab_index, True)
            self._set_tab_visible(self.points_tab_index, False)
            self._set_tab_visible(self.highlight_tab_index, False)
            self.tabs.setCurrentWidget(self.page_prompt)

    def _populate_reference_image_combo(self):
        current_id = str(self.reference_image_combo.currentData() or "").strip()
        image_files = []
        for file_obj in File.filter():
            data = file_obj.data if isinstance(getattr(file_obj, "data", None), dict) else {}
            if str(data.get("media_type", "")).strip().lower() != "image":
                continue
            image_files.append(file_obj)

        image_files.sort(key=lambda f: str((f.data or {}).get("name") or os.path.basename((f.data or {}).get("path", ""))).lower())
        for file_obj in image_files:
            data = file_obj.data if isinstance(file_obj.data, dict) else {}
            display_name = str(data.get("name") or os.path.basename(data.get("path", "")) or "Image").strip()
            icon = QIcon()
            thumb_path = GetThumbPath(file_obj.id, 1)
            if thumb_path and os.path.exists(thumb_path):
                icon = QIcon(thumb_path)
            self.reference_image_combo.addItem(icon, display_name, file_obj.id)
            self.reference_image_combo.setItemData(self.reference_image_combo.count() - 1, str(data.get("path", "")), Qt.ToolTipRole)

        if current_id:
            index = self.reference_image_combo.findData(current_id)
            if index >= 0:
                self.reference_image_combo.setCurrentIndex(index)

    def _choose_tracking_clicked(self):
        if not self.source_file:
            return

        win = SelectRegion(file=self.source_file, clip=None, selection_mode="annotate")
        if win.exec_() != QDialog.Accepted:
            return

        selection_payload = win.selection_payload()
        frame_size = win.videoPreview.curr_frame_size
        if not frame_size:
            frame_w = float(max(win.viewport_rect.width(), 1))
            frame_h = float(max(win.viewport_rect.height(), 1))
        else:
            frame_w = float(max(frame_size.width(), 1))
            frame_h = float(max(frame_size.height(), 1))
        src_w = float(max(getattr(win, "width", 1), 1))
        src_h = float(max(getattr(win, "height", 1), 1))

        def _scale_point_dict(p):
            if not isinstance(p, dict):
                return None
            try:
                x_in = float(p.get("x", 0.0))
                y_in = float(p.get("y", 0.0))
            except Exception:
                return None
            x_norm = max(min(x_in, float(max(frame_w - 1.0, 0.0))), 0.0)
            y_norm = max(min(y_in, float(max(frame_h - 1.0, 0.0))), 0.0)
            x_abs = int(round((x_norm / frame_w) * src_w))
            y_abs = int(round((y_norm / frame_h) * src_h))
            return {"x": x_abs, "y": y_abs}

        def _scale_rect_dict(r):
            if not isinstance(r, dict):
                return None
            try:
                x1_in = float(r.get("x1", 0.0))
                y1_in = float(r.get("y1", 0.0))
                x2_in = float(r.get("x2", 0.0))
                y2_in = float(r.get("y2", 0.0))
            except Exception:
                return None
            x1 = max(min(x1_in, float(max(frame_w - 1.0, 0.0))), 0.0)
            y1 = max(min(y1_in, float(max(frame_h - 1.0, 0.0))), 0.0)
            x2 = max(min(x2_in, float(max(frame_w - 1.0, 0.0))), 0.0)
            y2 = max(min(y2_in, float(max(frame_h - 1.0, 0.0))), 0.0)
            if x2 < x1:
                x1, x2 = x2, x1
            if y2 < y1:
                y1, y2 = y2, y1
            sx1 = int(round((x1 / frame_w) * src_w))
            sy1 = int(round((y1 / frame_h) * src_h))
            sx2 = int(round((x2 / frame_w) * src_w))
            sy2 = int(round((y2 / frame_h) * src_h))
            return {"x1": sx1, "y1": sy1, "x2": sx2, "y2": sy2}

        # Normalize all frame annotations to source frame coordinates.
        if isinstance(selection_payload, dict) and isinstance(selection_payload.get("frames"), dict):
            normalized_frames = {}
            for frame_key, frame_data in selection_payload.get("frames", {}).items():
                if not isinstance(frame_data, dict):
                    continue
                pos_pts = [_scale_point_dict(p) for p in (frame_data.get("positive_points") or [])]
                neg_pts = [_scale_point_dict(p) for p in (frame_data.get("negative_points") or [])]
                pos_rects = [_scale_rect_dict(r) for r in (frame_data.get("positive_rects") or [])]
                neg_rects = [_scale_rect_dict(r) for r in (frame_data.get("negative_rects") or [])]
                normalized_frames[str(frame_key)] = {
                    "positive_points": [p for p in pos_pts if p is not None],
                    "negative_points": [p for p in neg_pts if p is not None],
                    "positive_rects": [r for r in pos_rects if r is not None],
                    "negative_rects": [r for r in neg_rects if r is not None],
                }
            selection_payload["frames"] = normalized_frames

        frames = selection_payload.get("frames", {}) if isinstance(selection_payload, dict) else {}
        seed_frame = int(selection_payload.get("seed_frame", 1)) if isinstance(selection_payload, dict) else 1
        seed_data = frames.get(str(seed_frame), {}) if isinstance(frames, dict) else {}
        points_pos = list(seed_data.get("positive_points", []) or [])
        points_neg = list(seed_data.get("negative_points", []) or [])
        rects_pos = list(seed_data.get("positive_rects", []) or [])
        rects_neg = list(seed_data.get("negative_rects", []) or [])

        has_any_selection = bool(points_pos or points_neg or rects_pos or rects_neg)
        if not has_any_selection:
            QMessageBox.warning(
                self,
                "No Selections Found",
                "No points or rectangles were captured.",
            )
            return

        points_pos_text = json.dumps(points_pos) if points_pos else ""
        points_neg_text = json.dumps(points_neg) if points_neg else ""
        rects_pos_text = json.dumps(rects_pos) if rects_pos else ""
        rects_neg_text = json.dumps(rects_neg) if rects_neg else ""
        log.info(
            "Generate dialog captured SAM2 seed frame=%s points_pos=%s points_neg=%s rects_pos=%s rects_neg=%s",
            seed_frame,
            len(points_pos),
            len(points_neg),
            len(rects_pos),
            len(rects_neg),
        )
        self._coordinates_positive_text = points_pos_text
        self._coordinates_negative_text = points_neg_text
        self._rectangles_positive_text = rects_pos_text
        self._rectangles_negative_text = rects_neg_text
        self._auto_mode = False
        self._tracking_selection_payload = selection_payload if isinstance(selection_payload, dict) else {}
        self.points_preview.setPlainText(
            json.dumps(
                {
                    "seed_frame": seed_frame,
                    "auto_mode": False,
                    "positive": points_pos_text,
                    "negative": points_neg_text,
                    "positive_rects": rects_pos_text,
                    "negative_rects": rects_neg_text,
                    "tracking_selection": self._tracking_selection_payload,
                },
                indent=2,
            )
        )
        self.tabs.setCurrentWidget(self.page_points)

    def _clear_points_clicked(self):
        self._coordinates_positive_text = ""
        self._coordinates_negative_text = ""
        self._rectangles_positive_text = ""
        self._rectangles_negative_text = ""
        self._auto_mode = False
        self._tracking_selection_payload = {}
        self.points_preview.clear()

    def _set_tab_visible(self, index, visible):
        bar = self.tabs.tabBar()
        if hasattr(bar, "setTabVisible"):
            bar.setTabVisible(index, bool(visible))
        else:
            self.tabs.setTabEnabled(index, bool(visible))

    def _initialize_dialog_state(self):
        self._on_template_changed(self.template_combo.currentIndex())



"""
 @file
 @brief Painter for the timeline background gradient.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from PyQt5.QtCore import QRectF
from PyQt5.QtGui import QBrush, QColor, QLinearGradient, QPainter

from .base import BasePainter


class BackgroundPainter(BasePainter):
    def paint(self, painter: QPainter, rect: QRectF):
        bg = self.w.theme.background
        bg2 = getattr(self.w.theme, "background2", QColor())
        if bg2.isValid() and bg2 != bg:
            grad = QLinearGradient(rect.topLeft(), rect.bottomLeft())
            grad.setColorAt(0, bg)
            grad.setColorAt(1, bg2)
            painter.fillRect(rect, QBrush(grad))
        else:
            painter.fillRect(rect, bg)



"""Utilities for parsing and applying timeline CSS themes."""

import os
import re
from typing import Callable, Optional, Sequence, Tuple, Union

from PyQt5.QtCore import QFile, QByteArray
from PyQt5.QtGui import QColor, QPixmap
from classes.logger import log
from classes.info import PATH

LOG_THEME_MISS = False
LOG_THEME_INFO = False
# Verbose dump of parsed QWidget timeline theme values.
LOG_THEME_DUMP = False


def _apply_overrides(obj, overrides: dict, *, allow_unknown: bool = False) -> None:
    """Apply keyword overrides to a theme object, optionally ignoring unknown keys."""
    if not overrides:
        return
    allowed = set(obj.__dict__.keys())
    for key, value in overrides.items():
        if key not in allowed:
            if allow_unknown:
                continue
            raise TypeError("Unexpected theme option '%s'" % key)
        setattr(obj, key, value)


class BasicTheme:
    """Common style options for timeline elements."""

    def __init__(self, **kwargs):
        self.background: QColor = QColor()
        self.background2: QColor = QColor()
        self.border_color: QColor = QColor()
        self.border_radius: int = 0
        self.border_width: float = 0
        self.font_color: QColor = QColor()
        self.font_size: int = 0
        self.height: int = 0
        self.background_image: Optional[QPixmap] = None
        self.shadow_color: QColor = QColor()
        self.shadow_blur: int = 0
        self.thumb_width: int = 0
        self.thumb_height: int = 0
        self.top_overlay: QColor = QColor()
        self.top_overlay2: QColor = QColor()
        _apply_overrides(self, kwargs)


class TrackTheme(BasicTheme):
    """Theme for tracks."""

    def __init__(self, **kwargs):
        super().__init__()
        self.name_background: QColor = QColor()
        self.name_width: int = 0
        self.gap: int = 0
        self.margin_top: int = -1
        self.name_border_color: QColor = QColor()
        self.name_border_width: int = 0
        self.name_border_top_color: QColor = QColor()
        self.name_border_top_width: int = 0
        self.name_border_bottom_color: QColor = QColor()
        self.name_border_bottom_width: int = 0
        self.name_radius_tl: int = 0
        self.name_radius_bl: int = 0
        self.name_top_overlay: QColor = QColor()
        self.name_top_overlay2: QColor = QColor()
        _apply_overrides(self, kwargs)


def _color_repr(color: QColor) -> str:
    if not isinstance(color, QColor) or not color.isValid():
        return "invalid"
    return "#{:02X}{:02X}{:02X}{:02X}".format(
        color.red(),
        color.green(),
        color.blue(),
        color.alpha(),
    )


def _pixmap_repr(pixmap: Optional[QPixmap]) -> str:
    if not pixmap or pixmap.isNull():
        return "null"
    source = getattr(pixmap, "svg_path", None) or getattr(pixmap, "path", None)
    size = "{}x{}".format(pixmap.width(), pixmap.height())
    if source:
        return "pixmap {} ({})".format(size, source)
    return "pixmap {}".format(size)


def _theme_group_dump(title: str, values: Sequence[Tuple[str, str]]) -> None:
    lines = ["Theme Vars [{}]".format(title)]
    for key, value in values:
        lines.append("  {} = {}".format(key, value))
    log.info("\n".join(lines))


def _log_theme_dump(theme: "TimelineTheme", source: str) -> None:
    if not LOG_THEME_DUMP:
        return

    _theme_group_dump(
        "meta",
        [
            ("source", source),
        ],
    )
    _theme_group_dump(
        "timeline",
        [
            ("background", _color_repr(theme.background)),
            ("background2", _color_repr(theme.background2)),
            ("playhead_color", _color_repr(theme.playhead_color)),
            ("playhead_width", repr(theme.playhead_width)),
            ("clip_selected", _color_repr(theme.clip_selected)),
            ("selection", _color_repr(theme.selection)),
            ("selection_border", _color_repr(theme.selection_border)),
            ("selection_border_width", repr(theme.selection_border_width)),
            ("playback_cache_color", _color_repr(theme.playback_cache_color)),
            ("playback_cache_height", repr(theme.playback_cache_height)),
            ("ruler_name_background", _color_repr(theme.ruler_name_background)),
            ("ruler_name_background2", _color_repr(theme.ruler_name_background2)),
            ("ruler_time_font_size", repr(theme.ruler_time_font_size)),
            ("ruler_time_pad_left", repr(theme.ruler_time_pad_left)),
            ("ruler_time_pad_top", repr(theme.ruler_time_pad_top)),
            ("ruler_label_top", repr(theme.ruler_label_top)),
            ("scrollbar_handle", _color_repr(theme.scrollbar_handle)),
            ("scrollbar_track", _color_repr(theme.scrollbar_track)),
            ("scrollbar_width", repr(theme.scrollbar_width)),
            ("waveform_color", _color_repr(theme.waveform_color)),
            ("waveform_peak_color", _color_repr(theme.waveform_peak_color)),
            ("keyframe_fill", _color_repr(theme.keyframe_fill)),
            ("keyframe_border", _color_repr(theme.keyframe_border)),
            ("keyframe_inactive_opacity", repr(theme.keyframe_inactive_opacity)),
            ("keyframe_size", repr(theme.keyframe_size)),
            ("keyframe_panel_property_bg", _color_repr(theme.keyframe_panel_property_bg)),
            ("keyframe_panel_row_border_color", _color_repr(theme.keyframe_panel_row_border_color)),
            ("keyframe_panel_row_border_width", repr(theme.keyframe_panel_row_border_width)),
            ("keyframe_panel_curve_color", _color_repr(theme.keyframe_panel_curve_color)),
            ("keyframe_panel_marker_fill", _color_repr(theme.keyframe_panel_marker_fill)),
            ("keyframe_panel_marker_border", _color_repr(theme.keyframe_panel_marker_border)),
        ],
    )
    _theme_group_dump(
        "clip",
        [
            ("background", _color_repr(theme.clip.background)),
            ("background2", _color_repr(theme.clip.background2)),
            ("top_overlay", _color_repr(theme.clip.top_overlay)),
            ("top_overlay2", _color_repr(theme.clip.top_overlay2)),
            ("border_color", _color_repr(theme.clip.border_color)),
            ("border_radius", repr(theme.clip.border_radius)),
            ("border_width", repr(theme.clip.border_width)),
            ("font_color", _color_repr(theme.clip.font_color)),
            ("font_size", repr(theme.clip.font_size)),
            ("height", repr(theme.clip.height)),
            ("shadow_color", _color_repr(theme.clip.shadow_color)),
            ("shadow_blur", repr(theme.clip.shadow_blur)),
            ("thumb_width", repr(theme.clip.thumb_width)),
            ("thumb_height", repr(theme.clip.thumb_height)),
        ],
    )
    _theme_group_dump(
        "track",
        [
            ("background", _color_repr(theme.track.background)),
            ("background2", _color_repr(theme.track.background2)),
            ("border_color", _color_repr(theme.track.border_color)),
            ("border_radius", repr(theme.track.border_radius)),
            ("height", repr(theme.track.height)),
            ("gap", repr(theme.track.gap)),
            ("margin_top", repr(theme.track.margin_top)),
            ("font_color", _color_repr(theme.track.font_color)),
            ("font_size", repr(theme.track.font_size)),
            ("name_background", _color_repr(theme.track.name_background)),
            ("name_width", repr(theme.track.name_width)),
            ("name_border_color", _color_repr(theme.track.name_border_color)),
            ("name_border_width", repr(theme.track.name_border_width)),
            ("name_border_top_color", _color_repr(theme.track.name_border_top_color)),
            ("name_border_top_width", repr(theme.track.name_border_top_width)),
            ("name_border_bottom_color", _color_repr(theme.track.name_border_bottom_color)),
            ("name_border_bottom_width", repr(theme.track.name_border_bottom_width)),
            ("name_radius_tl", repr(theme.track.name_radius_tl)),
            ("name_radius_bl", repr(theme.track.name_radius_bl)),
            ("name_top_overlay", _color_repr(theme.track.name_top_overlay)),
            ("name_top_overlay2", _color_repr(theme.track.name_top_overlay2)),
        ],
    )
    _theme_group_dump(
        "transition",
        [
            ("background", _color_repr(theme.transition.background)),
            ("background2", _color_repr(theme.transition.background2)),
            ("top_overlay", _color_repr(theme.transition.top_overlay)),
            ("top_overlay2", _color_repr(theme.transition.top_overlay2)),
            ("border_color", _color_repr(theme.transition.border_color)),
            ("border_radius", repr(theme.transition.border_radius)),
            ("font_color", _color_repr(theme.transition.font_color)),
            ("font_size", repr(theme.transition.font_size)),
            ("height", repr(theme.transition.height)),
            ("background_image", _pixmap_repr(theme.transition.background_image)),
        ],
    )
    _theme_group_dump(
        "ruler",
        [
            ("background", _color_repr(theme.ruler.background)),
            ("background2", _color_repr(theme.ruler.background2)),
            ("border_color", _color_repr(theme.ruler.border_color)),
            ("font_color", _color_repr(theme.ruler.font_color)),
            ("font_size", repr(theme.ruler.font_size)),
            ("height", repr(theme.ruler.height)),
        ],
    )
    _theme_group_dump(
        "icons",
        [
            ("menu_icon", _pixmap_repr(theme.menu_icon)),
            ("menu_size", repr(theme.menu_size)),
            ("menu_margin", repr(theme.menu_margin)),
            ("playhead_icon", _pixmap_repr(theme.playhead_icon)),
            ("playhead_icon_width", repr(theme.playhead_icon_width)),
            ("playhead_icon_height", repr(theme.playhead_icon_height)),
            ("playhead_icon_offset_x", repr(theme.playhead_icon_offset_x)),
            ("playhead_icon_offset_y", repr(theme.playhead_icon_offset_y)),
            ("marker_icon", _pixmap_repr(theme.marker_icon)),
            ("marker_icon_width", repr(theme.marker_icon_width)),
            ("marker_icon_height", repr(theme.marker_icon_height)),
            ("marker_icon_offset_x", repr(theme.marker_icon_offset_x)),
            ("marker_icon_offset_y", repr(theme.marker_icon_offset_y)),
            ("track_keyframe_panel_disabled_icon", _pixmap_repr(theme.track_keyframe_panel_disabled_icon)),
            ("track_keyframe_panel_enabled_icon", _pixmap_repr(theme.track_keyframe_panel_enabled_icon)),
            ("keyframe_panel_add_icon", _pixmap_repr(theme.keyframe_panel_add_icon)),
            ("track_add_above_disabled_icon", _pixmap_repr(theme.track_add_above_disabled_icon)),
            ("track_add_above_enabled_icon", _pixmap_repr(theme.track_add_above_enabled_icon)),
            ("track_add_below_disabled_icon", _pixmap_repr(theme.track_add_below_disabled_icon)),
            ("track_add_below_enabled_icon", _pixmap_repr(theme.track_add_below_enabled_icon)),
            ("track_delete_disabled_icon", _pixmap_repr(theme.track_delete_disabled_icon)),
            ("track_delete_enabled_icon", _pixmap_repr(theme.track_delete_enabled_icon)),
            ("track_locked_disabled_icon", _pixmap_repr(theme.track_locked_disabled_icon)),
            ("track_locked_enabled_icon", _pixmap_repr(theme.track_locked_enabled_icon)),
            ("track_unlocked_disabled_icon", _pixmap_repr(theme.track_unlocked_disabled_icon)),
            ("track_unlocked_enabled_icon", _pixmap_repr(theme.track_unlocked_enabled_icon)),
        ],
    )


class TimelineTheme:
    """Container for all timeline related themes."""

    def __init__(self, **kwargs):
        self.background: QColor = QColor("#000")
        self.background2: QColor = QColor()
        self.playhead_color: QColor = QColor("#FFF")
        self.playhead_width: float = 0.0
        self.clip_selected: QColor = QColor("#FFF")
        self.selection: QColor = QColor(255, 255, 255, 80)
        self.selection_border: QColor = QColor()
        self.selection_border_width: float = 0.0
        self.playback_cache_color: QColor = QColor("#4B92AD")
        self.playback_cache_height: float = 5.0

        self.clip: BasicTheme = BasicTheme()
        self.transition: BasicTheme = BasicTheme()
        self.track: TrackTheme = TrackTheme()
        self.ruler: BasicTheme = BasicTheme()
        self.ruler_name_background: QColor = QColor()
        self.ruler_name_background2: QColor = QColor()
        self.ruler_time_font_size: int = 0
        self.menu_icon: Optional[QPixmap] = None
        self.menu_size: int = 0
        self.menu_margin: int = 0
        self.keyframe_toggle_off_icon: Optional[QPixmap] = None
        self.keyframe_toggle_on_icon: Optional[QPixmap] = None
        self.track_keyframe_panel_disabled_icon: Optional[QPixmap] = None
        self.track_keyframe_panel_enabled_icon: Optional[QPixmap] = None
        self.track_add_above_disabled_icon: Optional[QPixmap] = None
        self.track_add_above_enabled_icon: Optional[QPixmap] = None
        self.track_add_below_disabled_icon: Optional[QPixmap] = None
        self.track_add_below_enabled_icon: Optional[QPixmap] = None
        self.track_delete_disabled_icon: Optional[QPixmap] = None
        self.track_delete_enabled_icon: Optional[QPixmap] = None
        self.track_locked_disabled_icon: Optional[QPixmap] = None
        self.track_locked_enabled_icon: Optional[QPixmap] = None
        self.track_unlocked_disabled_icon: Optional[QPixmap] = None
        self.track_unlocked_enabled_icon: Optional[QPixmap] = None
        self.keyframe_panel_add_icon: Optional[QPixmap] = None
        self.keyframe_panel_property_bg: QColor = QColor()
        self.keyframe_panel_row_border_color: QColor = QColor()
        self.keyframe_panel_row_border_width: float = 1.0
        self.keyframe_panel_curve_color: QColor = QColor()
        self.keyframe_panel_marker_fill: QColor = QColor()
        self.keyframe_panel_marker_border: QColor = QColor()
        self.playhead_icon: Optional[QPixmap] = None
        self.playhead_icon_width: int = 0
        self.playhead_icon_height: int = 0
        self.playhead_icon_offset_x: int = 0
        self.playhead_icon_offset_y: int = 0
        self.marker_icon: Optional[QPixmap] = None
        self.marker_icon_width: int = 0
        self.marker_icon_height: int = 0
        self.marker_icon_offset_x: Optional[int] = None
        self.marker_icon_offset_y: Optional[int] = None
        self.marker_hit_padding: float = 4.0
        self.ruler_time_pad_left: int = 0
        self.ruler_time_pad_top: int = 0
        self.ruler_label_top: int = 0
        self.scrollbar_handle: QColor = QColor()
        self.scrollbar_track: QColor = QColor()
        self.scrollbar_width: int = 0
        self.waveform_color: QColor = QColor(42, 130, 218)
        self.waveform_peak_color: QColor = QColor(42, 130, 218, 128)
        self.keyframe_fill: QColor = QColor("#4d7bff")
        self.keyframe_border: QColor = QColor("#ffffff")
        self.keyframe_inactive_opacity: float = 0.5
        self.keyframe_size: int = 10
        _apply_overrides(self, kwargs)


DEFAULT_THEME = TimelineTheme()

# Load the main timeline CSS used by the web backends. Many timeline style
# values are defined here and are reused by the QWidget backend.
_CSS_PATH = os.path.normpath(os.path.join(
    os.path.dirname(__file__),
    "../../..",
    "timeline/media/css/main.css",
))
try:
    with open(_CSS_PATH, "r", encoding="utf-8") as _f:
        MAIN_CSS = _f.read()
except OSError:
    MAIN_CSS = ""


def _annotate_svg_metadata(pix: Optional[QPixmap], source: Optional[str]) -> Optional[QPixmap]:
    """Attach SVG metadata (path + bytes) to *pix* when available."""
    if not pix or pix.isNull() or not source or not source.lower().endswith(".svg"):
        return pix
    try:
        pix.svg_path = source
        data: Optional[bytes] = None
        if source.startswith(":"):
            file_obj = QFile(source)
            if file_obj.open(QFile.ReadOnly):
                try:
                    data = bytes(file_obj.readAll())
                finally:
                    file_obj.close()
        else:
            with open(source, "rb") as fh:
                data = fh.read()
        if data:
            pix.svg_bytes = data
            try:
                pix.svg_qbytearray = QByteArray(data)
            except Exception:
                pass
    except Exception:
        pass
    return pix


def _load_pixmap_with_meta(path: str) -> Optional[QPixmap]:
    pix = QPixmap(path)
    if pix.isNull():
        return None
    return _annotate_svg_metadata(pix, path)


def _css_prop(
    css: str,
    selector: str,
    prop: str,
    source: str,
    *,
    log_selector: bool = LOG_THEME_MISS,
    log_property: bool = LOG_THEME_MISS,
) -> Optional[str]:
    """Return property *prop* from the CSS *selector* block.

    Logging can be disabled for selector or property misses using the optional
    flags. This is useful when calling code plans to fall back to alternate
    property names and does not want intermediate MISS messages.
    """
    block_pat = rf"{re.escape(selector)}\s*\{{([^}}]*)\}}"
    m = re.search(block_pat, css, re.MULTILINE)
    if not m:
        if log_selector:
            log.info("Theme MISS [%s] selector '%s'", source, selector)
        return None
    block = m.group(1)
    m2 = re.search(rf"(?:^|;)\s*{re.escape(prop)}\s*:\s*([^;]+)", block)
    if not m2:
        if log_property:
            log.info(
                "Theme MISS [%s] selector '%s' property '%s'",
                source,
                selector,
                prop,
            )
        return None
    return m2.group(1).strip()


def _color_from_str(val: str) -> Optional[QColor]:
    """Parse a CSS color value into a ``QColor``.

    Supports hex colors and ``rgb/rgba`` declarations with either integer or
    float components. Returns ``None`` if the string cannot be parsed.
    """
    val = val.strip()
    if not val:
        return None
    if val.startswith("#"):
        col = QColor(val)
        return col if col.isValid() else None
    m = re.match(r"rgba?\(([^)]+)\)", val)
    if m:
        parts = [p.strip() for p in m.group(1).split(",")]
        if len(parts) >= 3:
            try:
                r = int(float(parts[0]))
                g = int(float(parts[1]))
                b = int(float(parts[2]))
                a = 255
                if len(parts) >= 4:
                    a_part = parts[3]
                    if a_part.endswith("%"):
                        a = int(float(a_part[:-1]) * 2.55)
                    else:
                        fa = float(a_part)
                        a = int(fa * 255) if fa <= 1 else int(fa)
                return QColor(r, g, b, a)
            except ValueError:
                return None
    col = QColor(val)
    return col if col.isValid() else None


def _parse_color(
    css: str,
    selector: str,
    prop: Union[str, Sequence[str]],
    source: str,
    *,
    log_miss: bool = LOG_THEME_MISS,
    log_selector: bool = LOG_THEME_MISS,
) -> Optional[QColor]:
    props = (prop,) if isinstance(prop, str) else tuple(prop)
    val = None
    for i, p in enumerate(props):
        val = _css_prop(
            css,
            selector,
            p,
            source,
            log_selector=i == 0 and log_selector,
            log_property=False,
        )
        if val is not None:
            break
    if val is None:
        if log_miss:
            log.info(
                "Theme MISS [%s] selector '%s' property '%s'",
                source,
                selector,
                props[0],
            )
        return None
    m = re.search(r"#([0-9a-fA-F]{3,8})", val)
    if m:
        col = _color_from_str("#" + m.group(1))
        if col:
            return col
    m = re.search(r"rgba?\([^\)]+\)", val)
    if m:
        col = _color_from_str(m.group(0))
        if col:
            return col
    # Handle shorthand declarations like "1px solid red !important" by
    # scanning tokens from right to left and returning the first valid color.
    parts = re.split(r"\s+", val.strip())
    for token in reversed(parts):
        if token.lower() == "!important":
            continue
        col = _color_from_str(token)
        if col and col.isValid():
            return col
    if log_miss:
        log.info(
            "Theme MISS [%s] selector '%s' property '%s' invalid color '%s'",
            source,
            selector,
            prop,
            val,
        )
    return None


def _parse_gradient(
    css: str, selector: str, prop: str, source: str, *, log_miss: bool = LOG_THEME_MISS
):
    """Return up to two colors from a CSS gradient.

    The returned colors are ordered for a top-to-bottom gradient. If the CSS
    gradient specifies the opposite direction (bottom to top), the order of the
    colors is swapped so callers can simply paint from top to bottom.
    """
    val = _css_prop(css, selector, prop, source)
    if not val:
        if log_miss:
            log.info("Theme MISS [%s] selector '%s' property '%s'", source, selector, prop)
        return None, None
    cols = re.findall(r"#(?:[0-9a-fA-F]{3,8})|rgba?\([^\)]+\)", val)
    qcols = [_color_from_str(c) for c in cols]
    qcols = [c for c in qcols if c and c.isValid()]
    if not qcols:
        if log_miss:
            log.info(
                "Theme MISS [%s] selector '%s' property '%s' invalid gradient '%s'",
                source,
                selector,
                prop,
                val,
            )
        return None, None

    first = qcols[0]
    second = qcols[1] if len(qcols) > 1 else None

    # Detect bottom-to-top gradients and reverse the color order so callers can
    # always assume the first color is at the top.
    val_lower = val.lower()
    idx_bottom = val_lower.find("bottom")
    idx_top = val_lower.find("top")
    reverse = False
    if idx_bottom != -1 and idx_top != -1:
        reverse = idx_bottom < idx_top
    else:
        m = re.search(r"linear-gradient\((?:to\s+)?(top|bottom)", val_lower)
        if m:
            reverse = m.group(1) == "bottom"
        else:
            m = re.search(r"-webkit-linear-gradient\((top|bottom)", val_lower)
            if m:
                reverse = m.group(1) == "bottom"
    if reverse and second is not None:
        first, second = second, first

    return first, second


def _parse_float(
    css: str,
    selector: str,
    prop: Union[str, Sequence[str]],
    source: str,
    *,
    log_miss: bool = LOG_THEME_MISS,
    log_selector: bool = LOG_THEME_MISS,
) -> Optional[float]:
    props = (prop,) if isinstance(prop, str) else tuple(prop)
    val = None
    for i, p in enumerate(props):
        val = _css_prop(
            css,
            selector,
            p,
            source,
            log_selector=i == 0 and log_selector,
            log_property=False,
        )
        if val is not None:
            break
    if val is None:
        if log_miss:
            log.info(
                "Theme MISS [%s] selector '%s' property '%s'",
                source,
                selector,
                props[0],
            )
        return None
    m = re.search(r"(-?[0-9.]+)", val)
    if m:
        try:
            return float(m.group(1))
        except ValueError:
            pass
    if log_miss:
        log.info(
            "Theme MISS [%s] selector '%s' property '%s' invalid number '%s'",
            source,
            selector,
            props[0],
            val,
        )
    return None


def _parse_pixmap(
    css: str,
    selector: str,
    prop: str,
    source: str,
    *,
    log_miss: bool = LOG_THEME_MISS,
) -> Optional[QPixmap]:
    val = _css_prop(css, selector, prop, source)
    if not val:
        if log_miss:
            log.info("Theme MISS [%s] selector '%s' property '%s'", source, selector, prop)
        return None
    m = re.search(r"url\(([^)]+)\)", val)
    if m:
        path = m.group(1).strip('"\'')
        if path.startswith(":"):
            img = QPixmap(path)
            if not img.isNull():
                return _annotate_svg_metadata(img, path)
        if not os.path.isabs(path):
            base = os.path.dirname(_CSS_PATH)
            found = None
            for i in range(3):
                candidate = os.path.normpath(
                    os.path.join(base, *([".."] * i), path)
                )
                if os.path.exists(candidate):
                    found = candidate
                    break
            path = found or os.path.normpath(os.path.join(base, path))
        if os.path.exists(path):
            img = _load_pixmap_with_meta(path)
            if img:
                if LOG_THEME_INFO and selector in {".playhead-top", ".marker_icon"} and prop == "background-image":
                    log.info(
                        "Theme [%s] %s %s loaded '%s'",
                        source,
                        selector,
                        prop,
                        path,
                    )
                return img
    if log_miss:
        log.info(
            "Theme MISS [%s] selector '%s' property '%s' invalid pixmap '%s'",
            source,
            selector,
            prop,
            val,
        )
    return None


def _parse_box_shadow(
    css: str, selector: str, source: str, *, log_miss: bool = True
):
    """Return (color, blur) from a box-shadow property."""
    val = _css_prop(css, selector, "box-shadow", source)
    if not val:
        if log_miss:
            log.info(
                "Theme MISS [%s] selector '%s' property '%s'",
                source,
                selector,
                "box-shadow",
            )
        return None, None
    col = None
    m = re.search(r"#([0-9a-fA-F]{3,8})", val)
    if m:
        col = QColor("#" + m.group(1))
    else:
        m = re.search(r"rgba?\([^\)]+\)", val)
        if m:
            col = QColor(m.group(0))
    nums = re.findall(r"(-?[0-9.]+)", val)
    blur = int(float(nums[2])) if len(nums) >= 3 else None
    if col is None and blur is None and log_miss:
        log.info(
            "Theme MISS [%s] selector '%s' property '%s' invalid value '%s'",
            source,
            selector,
            "box-shadow",
            val,
        )
    return col, blur


def _theme_pixmap(
    qt_theme, selector: str, prop: str, *, log_miss: bool = LOG_THEME_MISS
) -> Optional[QPixmap]:
    if not qt_theme or not hasattr(qt_theme, "style_sheet"):
        return None
    val = _css_prop(qt_theme.style_sheet, selector, prop, "theme", log_selector=log_miss, log_property=log_miss)
    if not val:
        return None
    m = re.search(r"url\(([^)]+)\)", val)
    if m:
        path = m.group(1).strip('"\'')
        if path.startswith(":"):
            img = QPixmap(path)
            if not img.isNull():
                return _annotate_svg_metadata(img, path)
        module_path = os.path.dirname(__import__(qt_theme.__module__).__file__)
        if not os.path.isabs(path):
            module_parent = os.path.dirname(module_path)
            trimmed = path.lstrip("./")
            rel_path = path
            while rel_path.startswith("../"):
                rel_path = rel_path[3:]
            candidates = []
            bases = [module_path, module_parent, PATH, os.path.dirname(PATH)]
            for base in bases:
                if not base:
                    continue
                candidates.append(os.path.normpath(os.path.join(base, path)))
                if trimmed and trimmed != path:
                    candidates.append(os.path.normpath(os.path.join(base, trimmed)))
                if rel_path and rel_path != path:
                    candidates.append(os.path.normpath(os.path.join(base, rel_path)))
            seen = set()
            resolved = None
            for candidate in candidates:
                if candidate in seen:
                    continue
                seen.add(candidate)
                if os.path.exists(candidate):
                    resolved = candidate
                    break
            if resolved:
                path = resolved
            else:
                path = os.path.normpath(os.path.join(module_path, path))
        if os.path.exists(path):
            img = _load_pixmap_with_meta(path)
            if img:
                if LOG_THEME_INFO and selector in {".playhead-top", ".marker_icon"} and prop == "background-image":
                    log.info(
                        "Theme [theme] %s %s loaded '%s'",
                        selector,
                        prop,
                        path,
                    )
                return img
    if log_miss:
        log.info("Theme MISS [theme] %s %s invalid pixmap '%s'", selector, prop, val)
    return None


def _theme_get_color(
    qt_theme,
    selector: str,
    prop: Union[str, Sequence[str]],
    *,
    log_miss: bool = LOG_THEME_MISS,
):
    if not qt_theme:
        return None
    props = (prop,) if isinstance(prop, str) else tuple(prop)
    if hasattr(qt_theme, "get_color"):
        for p in props:
            col = qt_theme.get_color(selector, p)
            if col:
                return col
    if hasattr(qt_theme, "style_sheet"):
        return _parse_color(
            qt_theme.style_sheet,
            selector,
            props,
            "theme",
            log_miss=log_miss,
            log_selector=log_miss,
        )
    return None


def _theme_get_int(
    qt_theme,
    selector: str,
    prop: Union[str, Sequence[str]],
    *,
    log_miss: bool = LOG_THEME_MISS,
):
    if not qt_theme:
        return None
    props = (prop,) if isinstance(prop, str) else tuple(prop)
    if hasattr(qt_theme, "get_int"):
        for p in props:
            val = qt_theme.get_int(selector, p)
            if val is not None:
                return val
    if hasattr(qt_theme, "style_sheet"):
        val = _parse_float(
            qt_theme.style_sheet,
            selector,
            props,
            "theme",
            log_miss=log_miss,
            log_selector=log_miss,
        )
        if val is not None:
            return int(val)
    return None


def _assign_color(target, attr: str, color: Optional[QColor]) -> None:
    if color:
        setattr(target, attr, color)


def _assign_value(
    target,
    attr: str,
    value: Optional[Union[int, float]],
    *,
    transform: Optional[Callable[[Union[int, float]], Union[int, float]]] = None,
) -> None:
    if value is not None:
        if transform:
            value = transform(value)
        setattr(target, attr, value)


def _apply_gradient_with_fallback(
    target,
    attr_primary: str,
    attr_secondary: str,
    gradient_func: Callable[[], Tuple[Optional[QColor], Optional[QColor]]],
    fallback_func: Callable[[], Optional[QColor]],
    *,
    miss_log: Optional[Callable[[], None]] = None,
) -> None:
    col1, col2 = gradient_func()
    if col1:
        setattr(target, attr_primary, col1)
    if col2:
        setattr(target, attr_secondary, col2)
    elif col1:
        setattr(target, attr_secondary, QColor())
    if col1 is None and col2 is None:
        fallback = fallback_func()
        if fallback:
            setattr(target, attr_primary, fallback)
            setattr(target, attr_secondary, QColor())
        elif miss_log:
            miss_log()


def _theme_apply_color(
    target,
    attr: str,
    qt_theme,
    selector: str,
    prop: Union[str, Sequence[str]],
    *,
    log_miss: bool = LOG_THEME_MISS,
) -> None:
    col = _theme_get_color(qt_theme, selector, prop, log_miss=log_miss)
    _assign_color(target, attr, col)


def _theme_apply_int(
    target,
    attr: str,
    qt_theme,
    selector: str,
    prop: Union[str, Sequence[str]],
    *,
    transform: Optional[Callable[[Union[int, float]], Union[int, float]]] = None,
    log_miss: bool = LOG_THEME_MISS,
) -> None:
    val = _theme_get_int(qt_theme, selector, prop, log_miss=log_miss)
    _assign_value(target, attr, val, transform=transform)


def _theme_get_first_color(
    qt_theme,
    queries: Sequence[Sequence],
) -> Optional[QColor]:
    for query in queries:
        selector, prop = query[0], query[1]
        extra = query[2] if len(query) > 2 else {}
        col = _theme_get_color(qt_theme, selector, prop, **extra)
        if col:
            return col
    return None


def _set_default_if_missing(target, attrs: Sequence[str], value: int) -> None:
    for attr in attrs:
        if not getattr(target, attr):
            setattr(target, attr, value)


def _apply_css_color_value(
    target,
    attr: str,
    css: str,
    selector: str,
    prop: Union[str, Sequence[str]],
    source: str,
    *,
    log_miss: bool = LOG_THEME_MISS,
    log_selector: bool = LOG_THEME_MISS,
) -> None:
    col = _parse_color(
        css,
        selector,
        prop,
        source,
        log_miss=log_miss,
        log_selector=log_selector,
    )
    _assign_color(target, attr, col)


def _apply_css_float_value(
    target,
    attr: str,
    css: str,
    selector: str,
    prop: Union[str, Sequence[str]],
    source: str,
    *,
    log_miss: bool = LOG_THEME_MISS,
    log_selector: bool = LOG_THEME_MISS,
    transform: Optional[Callable[[float], Union[int, float]]] = None,
) -> None:
    val = _parse_float(
        css,
        selector,
        prop,
        source,
        log_miss=log_miss,
        log_selector=log_selector,
    )
    _assign_value(target, attr, val, transform=transform)


def _apply_clip_box_shadow(
    theme: TimelineTheme,
    css: str,
    source: str,
    *,
    log_miss: bool,
) -> None:
    val = _css_prop(css, ".clip", "box-shadow", source, log_selector=False, log_property=False)
    if not val:
        return
    col, blur = _parse_box_shadow(css, ".clip", source, log_miss=log_miss)
    if col:
        theme.clip.shadow_color = col
    if blur is not None:
        theme.clip.shadow_blur = blur


def _ruler_gradient(css: str, source: str) -> Tuple[Optional[QColor], Optional[QColor]]:
    col1, col2 = _parse_gradient(css, "#scrolling_ruler", "background", source, log_miss=False)
    if not col1 and not col2:
        col1, col2 = _parse_gradient(css, "#ruler", "background", source, log_miss=False)
    return col1, col2


def _ruler_theme_background(qt_theme) -> Optional[QColor]:
    col = _theme_get_color(
        qt_theme,
        "#scrolling_ruler",
        ("background", "background-color"),
        log_miss=False,
    )
    if not col:
        col = _theme_get_color(
            qt_theme,
            "#ruler",
            ("background", "background-color"),
            log_miss=False,
        )
    return col


def _ruler_css_background(css: str, source: str) -> Optional[QColor]:
    col = _parse_color(
        css,
        "#scrolling_ruler",
        ("background", "background-color"),
        source,
        log_miss=False,
    )
    if not col:
        col = _parse_color(
            css,
            "#ruler",
            ("background", "background-color"),
            source,
            log_miss=False,
        )
    return col


def _css_get_first_color(
    css: str,
    source: str,
    log_miss: bool,
    queries: Sequence[Sequence],
) -> Optional[QColor]:
    for query in queries:
        selector, prop = query[0], query[1]
        col = _parse_color(css, selector, prop, source, log_miss=log_miss)
        if col:
            return col
    return None

def _theme_apply_background(theme: TimelineTheme, qt_theme, css_sheet: str) -> None:
    col1, col2 = _parse_gradient(css_sheet, "body", "background", "theme", log_miss=False)
    if col1:
        theme.background = col1
    if col2:
        theme.background2 = col2
    elif col1:
        theme.background2 = QColor()
    if col1 is None and col2 is None:
        col = _theme_get_color(qt_theme, "body", ("background", "background-color"))
        if col:
            theme.background = col
            theme.background2 = QColor()


def _theme_apply_clip(theme: TimelineTheme, qt_theme, css_sheet: str) -> None:
    _apply_gradient_with_fallback(
        theme.clip,
        "background",
        "background2",
        lambda: _parse_gradient(css_sheet, ".clip", "background", "theme", log_miss=False),
        lambda: _theme_get_color(qt_theme, ".clip", ("background", "background-color")),
    )
    _theme_apply_color(theme.clip, "border_color", qt_theme, ".clip", ("border-top", "border"))
    _theme_apply_int(
        theme.clip,
        "border_width",
        qt_theme,
        ".clip",
        ("border-top", "border"),
        transform=float,
    )
    _theme_apply_int(theme.clip, "border_radius", qt_theme, ".clip", "border-radius")
    _theme_apply_int(theme.clip, "font_size", qt_theme, ".clip", "font-size")
    _theme_apply_color(theme.clip, "font_color", qt_theme, ".clip_label", "color")
    _theme_apply_int(theme.clip, "height", qt_theme, ".clip", "height")
    _apply_clip_box_shadow(theme, css_sheet, "theme", log_miss=False)
    _theme_apply_int(theme.clip, "thumb_width", qt_theme, ".thumb", "width")
    _theme_apply_int(theme.clip, "thumb_height", qt_theme, ".thumb", "height")
    _apply_gradient_with_fallback(
        theme.clip,
        "top_overlay",
        "top_overlay2",
        lambda: _parse_gradient(css_sheet, ".clip_top", "background", "theme", log_miss=False),
        lambda: _theme_get_color(qt_theme, ".clip_top", ("background", "background-color"), log_miss=False),
    )


def _theme_apply_selection(theme: TimelineTheme, qt_theme, css_sheet: str) -> None:
    col = _theme_get_color(qt_theme, ".ui-selected", ("border-top", "border"))
    if col:
        theme.clip_selected = col
    op = None
    if css_sheet:
        op = _parse_float(
            css_sheet,
            ".ui-selectable-helper",
            "opacity",
            "theme",
            log_miss=False,
            log_selector=False,
        )
    col = _theme_get_color(qt_theme, ".ui-selectable-helper", ("background", "background-color"))
    if col:
        if op is not None and col.alpha() == 255:
            col.setAlpha(int(255 * op))
        theme.selection = col
    col = _theme_get_color(qt_theme, ".ui-selectable-helper", ("border", "border-color"))
    if col:
        if op is not None and col.alpha() == 255:
            col.setAlpha(int(255 * op))
        theme.selection_border = col
    val = _theme_get_int(qt_theme, ".ui-selectable-helper", ("border", "border-width"))
    if val is not None:
        theme.selection_border_width = float(val)


def _theme_apply_transition(theme: TimelineTheme, qt_theme, css_sheet: str) -> None:
    col1, col2 = _parse_gradient(css_sheet, ".transition", "background", "theme", log_miss=False)
    if col1:
        theme.transition.background = col1
    if col2:
        theme.transition.background2 = col2
    elif col1:
        theme.transition.background2 = QColor()
    if col1 is None and col2 is None:
        col = _theme_get_color(qt_theme, ".transition", ("background", "background-color"))
        if col:
            theme.transition.background = col
            theme.transition.background2 = QColor()
    col = _theme_get_color(qt_theme, ".transition", ("border-top", "border"))
    if col:
        theme.transition.border_color = col
    val = _theme_get_int(qt_theme, ".transition", "border-radius")
    if val is not None:
        theme.transition.border_radius = val
    img = _theme_pixmap(qt_theme, ".transition", "background-image")
    if img:
        theme.transition.background_image = img
    col = _theme_get_color(qt_theme, ".transition_label", "color")
    if col:
        theme.transition.font_color = col
    val = _theme_get_int(qt_theme, ".transition", "font-size")
    if val is not None:
        theme.transition.font_size = val
    val = _theme_get_int(qt_theme, ".transition", "height")
    if val is not None:
        theme.transition.height = val


def _theme_apply_track(theme: TimelineTheme, qt_theme, css_sheet: str) -> None:
    _apply_gradient_with_fallback(
        theme.track,
        "background",
        "background2",
        lambda: _parse_gradient(css_sheet, ".track", "background", "theme", log_miss=False),
        lambda: _theme_get_color(qt_theme, ".track", ("background", "background-color")),
    )
    _theme_apply_color(theme.track, "border_color", qt_theme, ".track", ("border-top", "border"))
    _theme_apply_int(theme.track, "border_radius", qt_theme, ".track", "border-radius")
    col = _theme_get_first_color(
        qt_theme,
        [
            (".track_name", "color"),
            (".track_label", "color"),
        ],
    )
    _assign_color(theme.track, "font_color", col)
    _theme_apply_int(theme.track, "height", qt_theme, ".track", "height")
    _theme_apply_int(theme.track, "font_size", qt_theme, ".track_name", "font-size")
    _theme_apply_color(
        theme.track,
        "name_background",
        qt_theme,
        ".track_name",
        ("background", "background-color"),
    )
    _theme_apply_int(theme.track, "name_width", qt_theme, ".track_name", "width")
    _theme_apply_int(theme.track, "gap", qt_theme, ".track", "margin-bottom")
    _theme_apply_int(theme.track, "margin_top", qt_theme, ".track", "margin-top")
    for attr, prop in (
        ("name_border_color", "border-left"),
        ("name_border_top_color", ("border-top", "border")),
        ("name_border_bottom_color", ("border-bottom", "border")),
    ):
        _theme_apply_color(theme.track, attr, qt_theme, ".track_name", prop)
    for attr, prop in (
        ("name_border_width", "border-left"),
        ("name_border_top_width", ("border-top", "border")),
        ("name_border_bottom_width", ("border-bottom", "border")),
    ):
        _theme_apply_int(theme.track, attr, qt_theme, ".track_name", prop)
    _theme_apply_int(theme.track, "name_radius_tl", qt_theme, ".track_name", "border-top-left-radius")
    _theme_apply_int(
        theme.track,
        "name_radius_bl",
        qt_theme,
        ".track_name",
        "border-bottom-left-radius",
    )
    val = _theme_get_int(qt_theme, ".track_name", "border-radius")
    if val is not None:
        _set_default_if_missing(theme.track, ("name_radius_tl", "name_radius_bl"), val)
    _apply_gradient_with_fallback(
        theme.track,
        "name_top_overlay",
        "name_top_overlay2",
        lambda: _parse_gradient(css_sheet, ".track_top", "background", "theme", log_miss=False),
        lambda: _theme_get_color(
            qt_theme,
            ".track_top",
            ("background", "background-color"),
            log_miss=False,
        ),
    )


def _theme_apply_ruler(theme: TimelineTheme, qt_theme, css_sheet: str) -> None:
    _apply_gradient_with_fallback(
        theme.ruler,
        "background",
        "background2",
        lambda: _ruler_gradient(css_sheet, "theme"),
        lambda: _ruler_theme_background(qt_theme),
        miss_log=(
            (lambda: log.info(
                "Theme MISS [theme] selector '#scrolling_ruler' property 'background'"
            ))
            if LOG_THEME_MISS
            else None
        ),
    )
    _apply_gradient_with_fallback(
        theme,
        "ruler_name_background",
        "ruler_name_background2",
        lambda: _parse_gradient(css_sheet, "#ruler_label", "background", "theme", log_miss=False),
        lambda: _theme_get_color(qt_theme, "#ruler_label", ("background", "background-color")),
    )
    _theme_apply_color(theme.ruler, "border_color", qt_theme, ".tick_mark", "background-color")
    _theme_apply_color(theme.ruler, "font_color", qt_theme, "#ruler_time", "color")
    _theme_apply_int(theme, "ruler_time_font_size", qt_theme, "#ruler_time", "font-size")
    fs = _parse_float(css_sheet, ".ruler_time", "font-size", "theme", log_miss=False)
    if fs is not None:
        base = theme.ruler_time_font_size or 12
        theme.ruler.font_size = int(fs * base) if fs < 5 else int(fs)
    _theme_apply_int(theme, "ruler_label_top", qt_theme, ".ruler_time", "top")
    _theme_apply_int(theme.ruler, "height", qt_theme, "#ruler", "height")
    _theme_apply_int(theme, "ruler_time_pad_left", qt_theme, "#ruler_time", "padding-left")
    _theme_apply_int(theme, "ruler_time_pad_top", qt_theme, "#ruler_time", "padding-top")


def _theme_apply_playhead(theme: TimelineTheme, qt_theme) -> None:
    col = _theme_get_color(qt_theme, ".playhead-line", "background-color")
    if col:
        theme.playhead_color = col
    val = _theme_get_int(qt_theme, ".playhead-line", "width")
    if val is not None:
        theme.playhead_width = float(val)
    img = _theme_pixmap(qt_theme, ".playhead-top", "background-image")
    if img:
        theme.playhead_icon = img
    val = _theme_get_int(qt_theme, ".playhead-top", "width")
    if val is not None:
        theme.playhead_icon_width = val
    val = _theme_get_int(qt_theme, ".playhead-top", "height")
    if val is not None:
        theme.playhead_icon_height = val
    val = _theme_get_int(qt_theme, ".playhead-top", "margin-left")
    if val is not None:
        theme.playhead_icon_offset_x = val
    val = _theme_get_int(qt_theme, ".playhead-top", "margin-top")
    if val is not None:
        theme.playhead_icon_offset_y = val


def _theme_apply_markers(theme: TimelineTheme, qt_theme) -> None:
    img = _theme_pixmap(qt_theme, ".marker_icon", "background-image")
    if img:
        theme.marker_icon = img
    val = _theme_get_int(qt_theme, ".marker_icon", "width")
    if val is not None:
        theme.marker_icon_width = val
    val = _theme_get_int(qt_theme, ".marker_icon", "height")
    if val is not None:
        theme.marker_icon_height = val
    val = _theme_get_int(qt_theme, ".marker_icon", "margin-left")
    if val is not None:
        theme.marker_icon_offset_x = val
    val = _theme_get_int(qt_theme, ".marker_icon", "bottom")
    if val is not None:
        theme.marker_icon_offset_y = val


def _theme_apply_menu(theme: TimelineTheme, qt_theme) -> None:
    img = _theme_pixmap(qt_theme, ".menu", "background-image")
    if img:
        theme.menu_icon = img
    val = _theme_get_int(qt_theme, ".menu", "width")
    if val is not None:
        theme.menu_size = val
    val = _theme_get_int(qt_theme, ".menu", "margin")
    if val is not None:
        theme.menu_margin = val


def _theme_apply_keyframe_panel(theme: TimelineTheme, qt_theme) -> None:
    off_img = (
        _theme_pixmap(qt_theme, ".track-keyframe-panel-disabled", "background-image")
        or _theme_pixmap(qt_theme, ".keyframe-toggle-off", "background-image")
    )
    if off_img:
        theme.track_keyframe_panel_disabled_icon = off_img
        theme.keyframe_toggle_off_icon = off_img
    on_img = (
        _theme_pixmap(qt_theme, ".track-keyframe-panel-enabled", "background-image")
        or _theme_pixmap(qt_theme, ".keyframe-toggle-on", "background-image")
    )
    if on_img:
        theme.track_keyframe_panel_enabled_icon = on_img
        theme.keyframe_toggle_on_icon = on_img
    add_img = _theme_pixmap(qt_theme, ".keyframe-panel-add", "background-image")
    if add_img:
        theme.keyframe_panel_add_icon = add_img
    panel_bg = _theme_get_color(
        qt_theme,
        ".keyframe-panel-row",
        ("background", "background-color"),
        log_miss=False,
    )
    if not panel_bg:
        panel_bg = _theme_get_color(qt_theme, "QMenuBar", ("background", "background-color"))
    if panel_bg:
        theme.keyframe_panel_property_bg = panel_bg
    border_color = _theme_get_color(
        qt_theme,
        ".keyframe-panel-row",
        ("border", "border-color"),
        log_miss=False,
    )
    if border_color:
        theme.keyframe_panel_row_border_color = border_color
    border_width = _theme_get_int(
        qt_theme,
        ".keyframe-panel-row",
        ("border", "border-width"),
        log_miss=False,
    )
    if border_width is not None:
        theme.keyframe_panel_row_border_width = float(border_width)
    curve_color = _theme_get_color(qt_theme, ".keyframe-panel-curve", "color", log_miss=False)
    if curve_color:
        theme.keyframe_panel_curve_color = curve_color
    marker_fill = _theme_get_color(
        qt_theme,
        ".keyframe-panel-point",
        ("background", "background-color"),
        log_miss=False,
    )
    if marker_fill:
        theme.keyframe_panel_marker_fill = marker_fill
    marker_border = _theme_get_color(
        qt_theme,
        ".keyframe-panel-point",
        ("border", "border-color"),
        log_miss=False,
    )
    if marker_border:
        theme.keyframe_panel_marker_border = marker_border
    inactive = _parse_float(
        getattr(qt_theme, "style_sheet", ""),
        ".keyframe-panel-point",
        "opacity",
        "theme",
        log_miss=False,
        log_selector=False,
    )
    if inactive is not None:
        theme.keyframe_inactive_opacity = max(0.0, min(1.0, float(inactive)))


def _theme_apply_track_toolbar(theme: TimelineTheme, qt_theme) -> None:
    if not qt_theme:
        return

    def _set_icon(attr, selector):
        pix = _theme_pixmap(qt_theme, selector, "background-image")
        if pix:
            setattr(theme, attr, pix)

    _set_icon("track_add_above_disabled_icon", ".track-add-above-disabled")
    _set_icon("track_add_above_enabled_icon", ".track-add-above-enabled")
    _set_icon("track_add_below_disabled_icon", ".track-add-below-disabled")
    _set_icon("track_add_below_enabled_icon", ".track-add-below-enabled")
    _set_icon("track_delete_disabled_icon", ".track-delete-disabled")
    _set_icon("track_delete_enabled_icon", ".track-delete-enabled")
    _set_icon("track_locked_disabled_icon", ".track-locked-disabled")
    _set_icon("track_locked_enabled_icon", ".track-locked-enabled")
    _set_icon("track_unlocked_disabled_icon", ".track-unlocked-disabled")
    _set_icon("track_unlocked_enabled_icon", ".track-unlocked-enabled")


def _apply_theme_obj(theme: TimelineTheme, qt_theme) -> TimelineTheme:
    """Update *theme* from a Qt theme instance using BaseTheme helpers."""

    if not qt_theme:
        return theme

    css_sheet = getattr(qt_theme, "style_sheet", "")
    _theme_apply_background(theme, qt_theme, css_sheet)
    _theme_apply_clip(theme, qt_theme, css_sheet)
    _theme_apply_selection(theme, qt_theme, css_sheet)
    _theme_apply_transition(theme, qt_theme, css_sheet)
    _theme_apply_track(theme, qt_theme, css_sheet)
    _theme_apply_ruler(theme, qt_theme, css_sheet)
    _theme_apply_playhead(theme, qt_theme)
    _theme_apply_markers(theme, qt_theme)
    _theme_apply_menu(theme, qt_theme)
    _theme_apply_keyframe_panel(theme, qt_theme)
    _theme_apply_track_toolbar(theme, qt_theme)

    return theme


def _css_apply_background(theme: TimelineTheme, css: str, source: str, log_miss: bool) -> None:
    col1, col2 = _parse_gradient(css, "body", "background", source, log_miss=False)
    if col1:
        theme.background = col1
    if col2:
        theme.background2 = col2
    elif col1:
        theme.background2 = QColor()
    if col1 is None and col2 is None:
        col = _parse_color(css, "body", ("background", "background-color"), source, log_miss=log_miss)
        if col:
            theme.background = col
            theme.background2 = QColor()


def _css_apply_clip(theme: TimelineTheme, css: str, source: str, log_miss: bool) -> None:
    _apply_gradient_with_fallback(
        theme.clip,
        "background",
        "background2",
        lambda: _parse_gradient(css, ".clip", "background", source, log_miss=False),
        lambda: _parse_color(
            css,
            ".clip",
            ("background", "background-color"),
            source,
            log_miss=log_miss,
        ),
    )
    _apply_css_color_value(
        theme.clip,
        "border_color",
        css,
        ".clip",
        ("border-top", "border"),
        source,
        log_miss=False,
    )
    _apply_css_float_value(
        theme.clip,
        "border_width",
        css,
        ".clip",
        ("border-top", "border"),
        source,
        log_miss=False,
        transform=float,
    )
    _apply_css_float_value(
        theme.clip,
        "border_radius",
        css,
        ".clip",
        "border-radius",
        source,
        log_miss=False,
        transform=int,
    )
    _apply_css_color_value(theme.clip, "font_color", css, ".clip_label", "color", source, log_miss=log_miss)
    _apply_css_float_value(
        theme.clip,
        "height",
        css,
        ".clip",
        "height",
        source,
        log_miss=log_miss,
        transform=int,
    )
    _apply_clip_box_shadow(theme, css, source, log_miss=log_miss)
    _apply_css_float_value(
        theme.clip,
        "thumb_width",
        css,
        ".thumb",
        "width",
        source,
        log_miss=log_miss,
        transform=int,
    )
    _apply_css_float_value(
        theme.clip,
        "thumb_height",
        css,
        ".thumb",
        "height",
        source,
        log_miss=log_miss,
        transform=int,
    )
    _apply_css_float_value(
        theme.clip,
        "font_size",
        css,
        ".clip",
        "font-size",
        source,
        log_miss=log_miss,
        transform=int,
    )
    _apply_gradient_with_fallback(
        theme.clip,
        "top_overlay",
        "top_overlay2",
        lambda: _parse_gradient(css, ".clip_top", "background", source, log_miss=False),
        lambda: _parse_color(
            css,
            ".clip_top",
            ("background", "background-color"),
            source,
            log_miss=False,
        ),
    )


def _css_apply_selection(theme: TimelineTheme, css: str, source: str, log_miss: bool) -> None:
    col = _parse_color(css, ".ui-selected", ("border-top", "border"), source, log_miss=log_miss)
    if col:
        theme.clip_selected = col
    op = _parse_float(css, ".ui-selectable-helper", "opacity", source, log_miss=log_miss)
    col = _parse_color(css, ".ui-selectable-helper", ("background", "background-color"), source, log_miss=log_miss)
    if col:
        if op is not None and col.alpha() == 255:
            col.setAlpha(int(255 * op))
        theme.selection = col
    col = _parse_color(css, ".ui-selectable-helper", ("border", "border-color"), source, log_miss=log_miss)
    if col:
        if op is not None and col.alpha() == 255:
            col.setAlpha(int(255 * op))
        theme.selection_border = col
    val = _parse_float(css, ".ui-selectable-helper", ("border", "border-width"), source, log_miss=log_miss)
    if val is not None:
        theme.selection_border_width = float(val)


def _css_apply_transition(theme: TimelineTheme, css: str, source: str, log_miss: bool) -> None:
    col1, col2 = _parse_gradient(css, ".transition", "background", source, log_miss=False)
    if col1:
        theme.transition.background = col1
    if col2:
        theme.transition.background2 = col2
    elif col1:
        theme.transition.background2 = QColor()
    if col1 is None and col2 is None:
        col = _parse_color(css, ".transition", ("background", "background-color"), source, log_miss=log_miss)
        if col:
            theme.transition.background = col
            theme.transition.background2 = QColor()
    col = _parse_color(css, ".transition", ("border-top", "border"), source, log_miss=False)
    if col:
        theme.transition.border_color = col
    val = _parse_float(css, ".transition", "border-radius", source, log_miss=False)
    if val is not None:
        theme.transition.border_radius = int(val)
    img = _parse_pixmap(css, ".transition", "background-image", source, log_miss=log_miss)
    if img:
        theme.transition.background_image = img
    col = _parse_color(css, ".transition_label", "color", source, log_miss=log_miss)
    if col:
        theme.transition.font_color = col
    val = _parse_float(css, ".transition", "font-size", source, log_miss=log_miss)
    if val is not None:
        theme.transition.font_size = int(val)
    val = _parse_float(css, ".transition", "height", source, log_miss=log_miss)
    if val is not None:
        theme.transition.height = int(val)


def _css_apply_track(theme: TimelineTheme, css: str, source: str, log_miss: bool) -> None:
    _apply_gradient_with_fallback(
        theme.track,
        "background",
        "background2",
        lambda: _parse_gradient(css, ".track", "background", source, log_miss=False),
        lambda: _parse_color(
            css,
            ".track",
            ("background", "background-color"),
            source,
            log_miss=log_miss,
        ),
    )
    _apply_css_color_value(
        theme.track,
        "border_color",
        css,
        ".track",
        ("border-top", "border"),
        source,
        log_miss=False,
    )
    _apply_css_float_value(
        theme.track,
        "border_radius",
        css,
        ".track",
        "border-radius",
        source,
        log_miss=False,
        transform=int,
    )
    col = _css_get_first_color(
        css,
        source,
        log_miss,
        [
            (".track_name", "color"),
            (".track_label", "color"),
        ],
    )
    _assign_color(theme.track, "font_color", col)
    _apply_css_float_value(
        theme.track,
        "height",
        css,
        ".track",
        "height",
        source,
        log_miss=log_miss,
        transform=int,
    )
    _apply_css_color_value(
        theme.track,
        "name_background",
        css,
        ".track_name",
        ("background", "background-color"),
        source,
        log_miss=log_miss,
    )
    _apply_css_float_value(
        theme.track,
        "name_width",
        css,
        ".track_name",
        "width",
        source,
        log_miss=log_miss,
        transform=int,
    )
    _apply_css_float_value(
        theme.track,
        "gap",
        css,
        ".track",
        "margin-bottom",
        source,
        log_miss=log_miss,
        transform=int,
    )
    for attr, prop in (
        ("name_border_color", "border-left"),
        ("name_border_top_color", ("border-top", "border")),
        ("name_border_bottom_color", ("border-bottom", "border")),
    ):
        _apply_css_color_value(theme.track, attr, css, ".track_name", prop, source, log_miss=log_miss)
    for attr, prop in (
        ("name_border_width", "border-left"),
        ("name_border_top_width", ("border-top", "border")),
        ("name_border_bottom_width", ("border-bottom", "border")),
    ):
        _apply_css_float_value(
            theme.track,
            attr,
            css,
            ".track_name",
            prop,
            source,
            log_miss=log_miss,
            transform=int,
        )
    _apply_css_float_value(
        theme.track,
        "name_radius_tl",
        css,
        ".track_name",
        ("border-top-left-radius", "border-radius"),
        source,
        log_miss=log_miss,
        transform=int,
    )
    _apply_css_float_value(
        theme.track,
        "name_radius_bl",
        css,
        ".track_name",
        ("border-bottom-left-radius", "border-radius"),
        source,
        log_miss=log_miss,
        transform=int,
    )
    _apply_css_float_value(
        theme.track,
        "font_size",
        css,
        ".track_name",
        "font-size",
        source,
        log_miss=log_miss,
        transform=int,
    )
    _apply_gradient_with_fallback(
        theme.track,
        "name_top_overlay",
        "name_top_overlay2",
        lambda: _parse_gradient(css, ".track_top", "background", source, log_miss=False),
        lambda: _parse_color(
            css,
            ".track_top",
            ("background", "background-color"),
            source,
            log_miss=False,
        ),
    )


def _css_apply_ruler(theme: TimelineTheme, css: str, source: str, log_miss: bool) -> None:
    _apply_gradient_with_fallback(
        theme.ruler,
        "background",
        "background2",
        lambda: _ruler_gradient(css, source),
        lambda: _ruler_css_background(css, source),
        miss_log=(
            (lambda: log.info(
                "Theme MISS [%s] selector '#scrolling_ruler' property 'background'",
                source,
            ))
            if LOG_THEME_MISS and log_miss
            else None
        ),
    )
    _apply_gradient_with_fallback(
        theme,
        "ruler_name_background",
        "ruler_name_background2",
        lambda: _parse_gradient(css, "#ruler_label", "background", source, log_miss=log_miss),
        lambda: _parse_color(
            css,
            "#ruler_label",
            ("background", "background-color"),
            source,
            log_miss=log_miss,
        ),
    )
    _apply_css_color_value(
        theme.ruler,
        "border_color",
        css,
        ".tick_mark",
        "background-color",
        source,
        log_miss=log_miss,
    )
    _apply_css_color_value(
        theme.ruler,
        "font_color",
        css,
        "#ruler_time",
        "color",
        source,
        log_miss=log_miss,
    )
    _apply_css_float_value(
        theme,
        "ruler_time_font_size",
        css,
        "#ruler_time",
        "font-size",
        source,
        log_miss=log_miss,
        transform=int,
    )
    fs = _parse_float(css, ".ruler_time", "font-size", source, log_miss=log_miss)
    if fs is not None:
        base = theme.ruler_time_font_size or 12
        theme.ruler.font_size = int(fs * base) if fs < 5 else int(fs)
    _apply_css_float_value(
        theme,
        "ruler_label_top",
        css,
        ".ruler_time",
        "top",
        source,
        log_miss=log_miss,
        transform=int,
    )
    _apply_css_float_value(
        theme.ruler,
        "height",
        css,
        "#ruler",
        "height",
        source,
        log_miss=log_miss,
        transform=int,
    )
    _apply_css_float_value(
        theme,
        "ruler_time_pad_left",
        css,
        "#ruler_time",
        "padding-left",
        source,
        log_miss=log_miss,
        transform=int,
    )
    _apply_css_float_value(
        theme,
        "ruler_time_pad_top",
        css,
        "#ruler_time",
        "padding-top",
        source,
        log_miss=log_miss,
        transform=int,
    )


def _css_apply_playhead(theme: TimelineTheme, css: str, source: str, log_miss: bool) -> None:
    col = _parse_color(css, ".playhead-line", "background-color", source, log_miss=log_miss)
    if col:
        theme.playhead_color = col
    val = _parse_float(css, ".playhead-line", "width", source, log_miss=log_miss)
    if val is not None:
        theme.playhead_width = val
    img = _parse_pixmap(css, ".playhead-top", "background-image", source, log_miss=log_miss)
    if img:
        theme.playhead_icon = img
    val = _parse_float(css, ".playhead-top", "width", source, log_miss=log_miss)
    if val is not None:
        theme.playhead_icon_width = int(val)
    val = _parse_float(css, ".playhead-top", "height", source, log_miss=log_miss)
    if val is not None:
        theme.playhead_icon_height = int(val)
    val = _parse_float(css, ".playhead-top", "margin-left", source, log_miss=log_miss)
    if val is not None:
        theme.playhead_icon_offset_x = int(val)
    val = _parse_float(css, ".playhead-top", "margin-top", source, log_miss=log_miss)
    if val is not None:
        theme.playhead_icon_offset_y = int(val)


def _css_apply_markers(theme: TimelineTheme, css: str, source: str, log_miss: bool) -> None:
    img = _parse_pixmap(css, ".marker_icon", "background-image", source, log_miss=log_miss)
    if img:
        theme.marker_icon = img
    val = _parse_float(css, ".marker_icon", "width", source, log_miss=log_miss)
    if val is not None:
        theme.marker_icon_width = int(val)
    val = _parse_float(css, ".marker_icon", "height", source, log_miss=log_miss)
    if val is not None:
        theme.marker_icon_height = int(val)
    val = _parse_float(css, ".marker_icon", "margin-left", source, log_miss=log_miss)
    if val is not None:
        theme.marker_icon_offset_x = int(val)
    val = _parse_float(css, ".marker_icon", "bottom", source, log_miss=log_miss)
    if val is not None:
        theme.marker_icon_offset_y = int(val)


def _css_apply_menu(theme: TimelineTheme, css: str, source: str, log_miss: bool) -> None:
    img = _parse_pixmap(css, ".menu", "background-image", source, log_miss=log_miss)
    if img:
        theme.menu_icon = img
    val = _parse_float(css, ".menu", "width", source, log_miss=log_miss)
    if val is not None:
        theme.menu_size = int(val)
    margin = _css_prop(css, ".menu", "margin", source, log_selector=log_miss, log_property=log_miss)
    if margin:
        match = re.search(r"(-?[0-9.]+)", margin)
        if match:
            theme.menu_margin = int(float(match.group(1)))


def _css_apply_keyframe_panel(theme: TimelineTheme, css: str, source: str, log_miss: bool) -> None:
    off_img = _parse_pixmap(
        css,
        ".track-keyframe-panel-disabled",
        "background-image",
        source,
        log_miss=log_miss,
    ) or _parse_pixmap(css, ".keyframe-toggle-off", "background-image", source, log_miss=log_miss)
    if off_img:
        theme.track_keyframe_panel_disabled_icon = off_img
        theme.keyframe_toggle_off_icon = off_img
    on_img = _parse_pixmap(
        css,
        ".track-keyframe-panel-enabled",
        "background-image",
        source,
        log_miss=log_miss,
    ) or _parse_pixmap(css, ".keyframe-toggle-on", "background-image", source, log_miss=log_miss)
    if on_img:
        theme.track_keyframe_panel_enabled_icon = on_img
        theme.keyframe_toggle_on_icon = on_img
    add_img = _parse_pixmap(css, ".keyframe-panel-add", "background-image", source, log_miss=log_miss)
    if add_img:
        theme.keyframe_panel_add_icon = add_img
    panel_bg = _parse_color(
        css,
        ".keyframe-panel-row",
        ("background", "background-color"),
        source,
        log_miss=False,
    )
    if panel_bg:
        theme.keyframe_panel_property_bg = panel_bg
    border_color = _parse_color(
        css,
        ".keyframe-panel-row",
        ("border", "border-color"),
        source,
        log_miss=False,
    )
    if border_color:
        theme.keyframe_panel_row_border_color = border_color
    border_width = _parse_float(
        css,
        ".keyframe-panel-row",
        ("border", "border-width"),
        source,
        log_miss=False,
    )
    if border_width is not None:
        theme.keyframe_panel_row_border_width = float(border_width)
    curve_color = _parse_color(css, ".keyframe-panel-curve", "color", source, log_miss=False)
    if curve_color:
        theme.keyframe_panel_curve_color = curve_color
    marker_fill = _parse_color(
        css,
        ".keyframe-panel-point",
        ("background", "background-color"),
        source,
        log_miss=False,
    )
    if marker_fill:
        theme.keyframe_panel_marker_fill = marker_fill
    marker_border = _parse_color(
        css,
        ".keyframe-panel-point",
        ("border", "border-color"),
        source,
        log_miss=False,
    )
    if marker_border:
        theme.keyframe_panel_marker_border = marker_border
    inactive = _parse_float(
        css,
        ".keyframe-panel-point",
        "opacity",
        source,
        log_miss=False,
        log_selector=False,
    )
    if inactive is not None:
        theme.keyframe_inactive_opacity = max(0.0, min(1.0, float(inactive)))


def _css_apply_track_toolbar(theme: TimelineTheme, css: str, source: str, log_miss: bool) -> None:
    def _set_icon(attr, selector):
        pix = _parse_pixmap(css, selector, "background-image", source, log_miss=log_miss)
        if pix:
            setattr(theme, attr, pix)

    _set_icon("track_add_above_disabled_icon", ".track-add-above-disabled")
    _set_icon("track_add_above_enabled_icon", ".track-add-above-enabled")
    _set_icon("track_add_below_disabled_icon", ".track-add-below-disabled")
    _set_icon("track_add_below_enabled_icon", ".track-add-below-enabled")
    _set_icon("track_delete_disabled_icon", ".track-delete-disabled")
    _set_icon("track_delete_enabled_icon", ".track-delete-enabled")
    _set_icon("track_locked_disabled_icon", ".track-locked-disabled")
    _set_icon("track_locked_enabled_icon", ".track-locked-enabled")
    _set_icon("track_unlocked_disabled_icon", ".track-unlocked-disabled")
    _set_icon("track_unlocked_enabled_icon", ".track-unlocked-enabled")


def _css_apply_scrollbars(theme: TimelineTheme, css: str, source: str) -> None:
    col = _parse_color(css, "::-webkit-scrollbar-thumb", "background-color", source, log_miss=False)
    if col:
        theme.scrollbar_handle = col
    col = _parse_color(
        css,
        "::-webkit-scrollbar-track",
        ("background", "background-color"),
        source,
        log_miss=False,
    )
    if col:
        theme.scrollbar_track = col
    val = _parse_float(css, "::-webkit-scrollbar", ("width", "height"), source, log_miss=False)
    if val is not None:
        theme.scrollbar_width = int(val)


def _apply_css(theme: TimelineTheme, css: str, source: str = "css") -> TimelineTheme:
    """Update *theme* with values parsed from *css*."""

    if not css:
        return theme

    log_miss = False

    _css_apply_background(theme, css, source, log_miss)
    _css_apply_clip(theme, css, source, log_miss)
    _css_apply_selection(theme, css, source, log_miss)
    _css_apply_transition(theme, css, source, log_miss)
    _css_apply_track(theme, css, source, log_miss)
    _css_apply_ruler(theme, css, source, log_miss)
    _css_apply_playhead(theme, css, source, log_miss)
    _css_apply_markers(theme, css, source, log_miss)
    _css_apply_menu(theme, css, source, log_miss)
    _css_apply_keyframe_panel(theme, css, source, log_miss)
    _css_apply_track_toolbar(theme, css, source, log_miss)
    _css_apply_scrollbars(theme, css, source)

    return theme

def apply_theme(widget, css: str = "") -> bool:
    """Load theme values for *widget* and return True if geometry changed."""

    from classes.app import get_app

    app_theme = get_app().theme_manager.get_current_theme() if get_app() else None
    theme_name = getattr(app_theme, "name", "") if app_theme else ""

    t = TimelineTheme()

    # Start with defaults from the main CSS file
    t = _apply_css(t, MAIN_CSS, source="main.css")

    # Override with values from the active Qt theme instance
    if app_theme:
        t = _apply_theme_obj(t, app_theme)

    # Optional additional CSS overrides
    if isinstance(css, str) and css.strip():
        t = _apply_css(t, css, source="override")

    normalized_theme_name = str(theme_name or "").strip().lower()
    normalized_theme_name = normalized_theme_name.replace(":", "").replace("-", " ")
    normalized_theme_name = " ".join(normalized_theme_name.split())
    theme_class_name = (
        app_theme.__class__.__name__.strip().lower() if app_theme else ""
    )
    is_humanity_dark = (
        normalized_theme_name == "humanity dark"
        or theme_class_name == "humanitydarktheme"
    )

    if not t.playhead_icon:
        base = os.path.dirname(_CSS_PATH)
        default_path = os.path.normpath(os.path.join(base, "../images/playhead.svg"))
        if os.path.exists(default_path):
            t.playhead_icon = QPixmap(default_path)
            if LOG_THEME_INFO:
                log.info(
                    "Theme [default] .playhead-top background-image loaded '%s'",
                    default_path,
                )

    def _load_fallback_icon(*parts):
        path = os.path.normpath(os.path.join(PATH, *parts))
        if not os.path.exists(path):
            return None
        pix = _load_pixmap_with_meta(path)
        if not pix:
            return None
        if LOG_THEME_INFO:
            log.info("Theme [default] fallback icon loaded '%s'", path)
        return pix

    def _ensure_icon(attr, parts):
        pix = getattr(t, attr, None)
        if pix and not pix.isNull():
            return
        fallback = _load_fallback_icon(*parts)
        if fallback:
            setattr(t, attr, fallback)

    track_keyframe_disabled = (
        "humanity-dark-track-keyframe-panel-show-disabled.svg"
        if is_humanity_dark
        else "track-keyframe-panel-show-disabled.svg"
    )
    track_keyframe_enabled = (
        "humanity-dark-track-keyframe-panel-show-enabled.svg"
        if is_humanity_dark
        else "track-keyframe-panel-show-enabled.svg"
    )
    track_locked_disabled = (
        "humanity-dark-track-locked-disabled.svg"
        if is_humanity_dark
        else "track-locked-disabled.svg"
    )
    track_locked_enabled = (
        "humanity-dark-track-locked-enabled.svg"
        if is_humanity_dark
        else "track-locked-enabled.svg"
    )
    track_unlocked_disabled = (
        "humanity-dark-track-unlocked-disabled.svg"
        if is_humanity_dark
        else "track-unlocked-disabled.svg"
    )
    track_unlocked_enabled = (
        "humanity-dark-track-unlocked-enabled.svg"
        if is_humanity_dark
        else "track-unlocked-enabled.svg"
    )

    _fallback_map = {
        "track_keyframe_panel_disabled_icon": ("themes", "humanity", "images", track_keyframe_disabled),
        "track_keyframe_panel_enabled_icon": ("themes", "humanity", "images", track_keyframe_enabled),
        "track_add_above_disabled_icon": ("themes", "humanity", "images", "track-add-above-disabled.svg"),
        "track_add_above_enabled_icon": ("themes", "humanity", "images", "track-add-above-enabled.svg"),
        "track_add_below_disabled_icon": ("themes", "humanity", "images", "track-add-below-disabled.svg"),
        "track_add_below_enabled_icon": ("themes", "humanity", "images", "track-add-below-enabled.svg"),
        "track_delete_disabled_icon": ("themes", "humanity", "images", "track-delete-disabled.svg"),
        "track_delete_enabled_icon": ("themes", "humanity", "images", "track-delete-enabled.svg"),
        "track_locked_disabled_icon": ("themes", "humanity", "images", track_locked_disabled),
        "track_locked_enabled_icon": ("themes", "humanity", "images", track_locked_enabled),
        "track_unlocked_disabled_icon": ("themes", "humanity", "images", track_unlocked_disabled),
        "track_unlocked_enabled_icon": ("themes", "humanity", "images", track_unlocked_enabled),
        "marker_icon": ("timeline", "media", "images", "markers", "marker.svg"),
    }

    for attr, parts in _fallback_map.items():
        _ensure_icon(attr, parts)

    if t.marker_icon and not t.marker_icon.isNull():
        if not t.marker_icon_width:
            t.marker_icon_width = t.marker_icon.width()
        if not t.marker_icon_height:
            t.marker_icon_height = t.marker_icon.height()

    if (not t.keyframe_toggle_off_icon or t.keyframe_toggle_off_icon.isNull()) and t.track_keyframe_panel_disabled_icon:
        t.keyframe_toggle_off_icon = t.track_keyframe_panel_disabled_icon
    if (not t.keyframe_toggle_on_icon or t.keyframe_toggle_on_icon.isNull()) and t.track_keyframe_panel_enabled_icon:
        t.keyframe_toggle_on_icon = t.track_keyframe_panel_enabled_icon

    if not t.keyframe_panel_add_icon or t.keyframe_panel_add_icon.isNull():
        fallback = _load_fallback_icon("themes", "humanity", "images", "keyframe-panel-add.svg")
        if fallback:
            t.keyframe_panel_add_icon = fallback

    if not t.menu_icon or t.menu_icon.isNull():
        fallback = _load_fallback_icon("timeline", "media", "images", "menu.svg")
        if fallback:
            t.menu_icon = fallback
            if not t.menu_size:
                t.menu_size = fallback.width()

    # Limit recent QWidget parity tweaks to Humanity Dark only.
    if not is_humanity_dark:
        t.clip.top_overlay = QColor()
        t.clip.top_overlay2 = QColor()
        t.track.name_top_overlay = QColor()
        t.track.name_top_overlay2 = QColor()
        t.track.name_radius_tl = 0
        t.track.name_radius_bl = 0

    theme_name_for_log = theme_name or "none"
    sources = ["main.css"]
    if app_theme:
        sources.append("qt-theme:{}".format(theme_name_for_log))
    if isinstance(css, str) and css.strip():
        sources.append("override-css")
    _log_theme_dump(t, " + ".join(sources))


    old_track_h = widget.track_height
    old_name_w = widget.track_name_width
    old_ruler_h = widget.ruler_height
    old_gap = getattr(widget, 'track_gap', 0)
    old_margin_top = getattr(widget, 'track_margin_top', 0)

    widget.theme = t

    if t.track.height:
        widget.track_height = t.track.height
    if t.track.name_width:
        widget.track_name_width = t.track.name_width
    if t.track.gap:
        widget.track_gap = t.track.gap
    if t.track.margin_top >= 0:
        widget.track_margin_top = t.track.margin_top
    else:
        current_gap = getattr(widget, 'track_gap', 0)
        if current_gap:
            widget.track_margin_top = current_gap
    if t.ruler.height:
        widget.ruler_height = t.ruler.height
    if t.scrollbar_width:
        widget.scroll_bar_thickness = t.scrollbar_width

    return (
        old_track_h != widget.track_height
        or old_name_w != widget.track_name_width
        or old_ruler_h != widget.ruler_height
        or old_gap != widget.track_gap
        or old_margin_top != getattr(widget, 'track_margin_top', 0)
    )
